#ifndef CRFController_hh
#define CRFController_hh

//#include "CRFBasic.hh"

using namespace tree_mcmc;

// $Id$
namespace libpg {

/** 
 *	The Conditional Random Field (CRF) controller which is NOT built upon any approximator.
 *	This class now deliberately uses global variables to communicate with the tree sampler.
 *	In later version, a tree sampler object will become a member of CRFController.
 *	So reading/writing tree sampler variables will be similar to global variables.
 */

class CRFController : public Controller {

public:
   
    /**
	 *	variables
	 */

    /**
	 * Number of observations/inputs at each agent, assuming it is equal for all agents.
	 * E.g., there are 2 observations at each intersection: horizontal & vertical
	 * Each observation can be continuous real valued.
	 */ 
	unsigned int numObs;

	/** 
	 * Number of action at each node (assuming each node has same # of action)
	 * Actions must be discrete and finite.
	 */
	unsigned int numAction;		

	/************************************************************************/
	/*  Notice: trace, exp (grad), para are served by Tree_MCMC.cpp			*/
	/************************************************************************/
    
    /**
     * Discount factor
     */
    double discount;
    
    /**
     * Step sizes
     */
    double step_size[10];
	double ori_step_size[10];

	
	double maxEdgeRand, maxNodeRand;	// range of random initialization of features

	CRFBasic *crf_engine;
	int app_id;
     	   
public:

    /**
	 *	Methods
	 */        	
	CRFController(int *argc, char** argv);
    virtual ~CRFController() {};	// destructor

    /**
     *  Where the real coding is done to implement 
     */
    virtual void getAction(Observation& obs, Vector& action, bool computeGrad = true);
    virtual void getMostProbAction(Observation& obs, Vector& action);

    virtual int	 getInputDim();
    virtual int	 getNumParams();
    virtual void discountTrace();
    virtual void setDiscount(double discount);
    virtual void accumulateGrad(Vector& reward, Observation& newObs);
    virtual void resetGrad();
    virtual void instantStep(Vector& reward);
    virtual void batchStep();
    virtual void computeDirection(int steps);
    virtual void setStepSize(double stepSize);
    virtual void resetTrace();
    virtual void resetParams();
    virtual void randomizeParams(double maxRand);
    virtual double getMaxParam();
    virtual void write(std::ostream& o);
    virtual void read(std::istream& o);

    virtual void reduce(Vector& v, Approximator::StatsEnum s);
    virtual void scatter(Vector& v, Approximator::StatsEnum s);

	virtual void buffer_last_action(int * actions)  {crf_engine->buffer_last_action(actions);};

	virtual char * get_diagnosis(int argc, char ** argv);
	virtual int get_num_agent() {return crf_engine->get_num_agent();};
	virtual int get_num_edge() {return crf_engine->get_num_edge();};
	virtual bool distribute_reward() { return crf_engine->distribute_reward();};
	virtual void fill_edge_reward(double *rewards) {crf_engine->fill_edge_reward(rewards);};

	int get_num_node_extra_fea() {return crf_engine->get_num_node_extra_fea();};
	int get_num_class() {return crf_engine->get_num_class();};
	int get_num_edge_fea() {return crf_engine->get_num_edge_fea();};

	virtual bool is_independent() {return false;};


private:
	void continue_fea_para(char *filename);
};
}
#endif
