#ifndef ManyToOneController_hh
#define ManyToOneController_hh
// $Id: ManyToOneController.hh 29 2006-12-19 02:16:20Z daa $

namespace libpg {

/**
 * This controller is a collection of Approximators
 * Each approximator is equivalent to a single action
 * The softmax calculation is performed over all approximator 
 * outputs.
 */

class ManyToOneController : public Controller {


public:

    /**
     * the sub controllers
     */
    typedef std::vector<Approximator*> Approximators;
    Approximators approximators;

    
    bool splitObs; // Does each controller get its own observation
    bool localRewards; // Is there a reward per controller
    
    Vector dummyDist;
    Vector dist;
    

    ManyToOneController(Approximators approximators, bool splitObs, bool localRewards);
    virtual ~ManyToOneController() {};

    /**
     * Where the real coding is done to implement 
     */
    virtual void getAction(Observation& obs, Vector& action, bool computeGrad = true);
    
    virtual void getMostProbAction(Observation& obs, Vector& action); 

    virtual int getInputDim();

    virtual void discountTrace();
    virtual void setDiscount(double discount);
    virtual void accumulateGrad(Vector& rewards, Observation& newObs);
    virtual void resetGrad();
    virtual void instantStep(Vector& rewards);
    virtual void computeDirection(int steps);
    virtual void batchStep();
    virtual void setStepSize(double stepSize);
    virtual void resetTrace();
    virtual void resetParams();
    virtual void randomizeParams(double maxRand);
    virtual double getMaxParam();
    virtual int getNumParams();
    virtual void write(std::ostream& o);
    virtual void read(std::istream& o);

};
}
#endif
