#include"PGBasics.hh"
#include"BasicController.hh"
#include"ManyToOneController.hh"
#include"ManyToOneControllerMinUpdates.hh"

namespace libpg {

ManyToOneControllerMinUpdates::ManyToOneControllerMinUpdates(Approximators approximators, bool splitObs, bool localRewards)
    : ManyToOneController::ManyToOneController(approximators, splitObs, localRewards) {

    lastUpdateTime.resize(approximators.size());

    currentTime = 0;
    lastUpdateTime.clear();    
}


void ManyToOneControllerMinUpdates::getAction(Observation& obs, Vector& action, bool computeGrad) {
    ManyToOneController::getAction( obs, action, computeGrad);
    eligible = &(obs.getEligible());
}


/**
 *
 * The remaining methods just pass onto equiv Approximator
 * functions. We have to do it in this ugly way because more advanced
 * multiple agent controllers will need to override these functions to
 * update all agents in turn.
 *
 */
void ManyToOneControllerMinUpdates::discountTrace() {

    currentTime++;

    int j=0;
    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	if ((*eligible)(j)) {
	    (*i)->setDiscount(pow(discount, currentTime - lastUpdateTime(j)));
	    (*i)->discountTrace();
	    lastUpdateTime(j) = currentTime;
	}
	j++;
    }
}


void ManyToOneControllerMinUpdates::setDiscount(double discount) {
    this->discount = discount;
}


void ManyToOneControllerMinUpdates::instantStep(Vector& rewards) {

    double reward;
    assert((localRewards && rewards.size()==approximators.size()) || 
	  (!localRewards && rewards.size()==1));

    for (uint i = 0; i < approximators.size(); i++) {
	reward = localRewards ? rewards[i] : rewards[0];
	if (reward != 0.0) {
	    if (!(*eligible)(i)) {
		// If action eligible, trace has already been updated
		approximators[i]->setDiscount(pow(discount, currentTime - lastUpdateTime(i)));
		approximators[i]->discountTrace();
		lastUpdateTime(i) = currentTime;
	    }
	    approximators[i]->instantStep(reward);
	}
    }
}

void ManyToOneControllerMinUpdates::resetTrace() {
    ManyToOneController::resetTrace();

    currentTime = 0;
    lastUpdateTime.clear();
}

}
