#ifndef Observation_hh
#define Observation_hh
// $Id: Observation.hh 29 2006-12-19 02:16:20Z daa $

namespace libpg {
class Observation {

protected:



    /**
     * The actual full observation, filled in by
     * the simulator
     */
    Matrix features;
    
    /**
     * A vector that describes which action are available It's up to a
     * controller to decide how to use this, if at all. The inital
     * dimension is 0 which can be tested by controllers who require
     * this vector using an assert
     */
    Vector eligible;
    
    /**
     * Another optional variable designed for multiple controllers
     * where the observation for each agent is encoded as a row
     * or column of the observation vector. Determines which
     * row or col should be used in the current controller.
     * Basically an efficiency thing, avoiding copying a 
     * potentially large observation column into a stand alone
     * column
     */
    int agent; 

    /**
     * A count of the number of steps in optimisation/evaluation so
     * far.  This may only be reset in between epochs (in the case of
     * batch learning) Or after a complete learning reset (i.e., reset
     * of parameters).  Any part of the controller/agent code may use
     * this. E.g., if the policy is non-stationary, or if step sizes
     * need to be computed based on optimisation steps.  It may only
     * be set by the RLAlg class.
     */
   int steps;

public:

    Observation(int rows, int cols, int agents);

    Observation(Observation& obs);
    Observation() {}; // Allow creation of a blank Observation

    /**
     * Resize the observation matricies and vectors.
     * Normally done by the constructor with the same arguments.
     */
    void init(int rows, int cols, int agents);
    
    /**
     * Get count of the number of steps in optimisation/evaluation so
     * far.  This may only be reset in between epochs (in the case of
     * batch learning) Or after a complete learning reset (i.e., reset
     * of parameters).  Any part of the controller/agent code may use
     * this. E.g., if the policy is non-stationary, or if step sizes
     * need to be computed based on optimisation steps.  It may only
     * be set by the RLAlg class.
     */
    int getSteps() { return steps; }

    /**
     * Set the number of steps. Only RLAlg should call this.
     * @param s new step number
     */
    void setSteps(int s) { steps = s; }


    /**
     * set which agent this observation is destined for. Used
     * by approximators and controllers that single out a particular
     * part of the feature matrix depending on the agent.
     */
    void setAgent(int a) { agent = a; }

    /**
     * Find out which agent is being represented now. Assists with
     * pulling out the correct part of the features, but exactly
     * how this is done depends on the controller and approximators 
     * involved.
     */
    int getAgent() { return agent; }

    /**
     * Get the observation matrix.
     */
    Matrix& getFeatures() { return features; }

    /**
     * Get the eligibility vector
     */
    Vector& getEligible() { return eligible; }
    
    /**
     * Perfect copy of observation, including resizing
     */
   Observation& operator=(Observation& rhs);

};
}
#endif
