#ifndef SARSAController_hh
#define SARSAController_hh

namespace libpg {
/**
 * This is a subclass of BasicController, implementing a SARSA specific
 * controller. The action selection rule used is softmax with Gibbs
 * (or Boltzmann) distribution.
 */

class SARSAController : public BasicController {


private:

    int lastAct;         // Last action taken
    Observation lastObs; // Last observation
    double error;        // Temporal difference error
    double lastActVal;   // Last action value
    double discount;     // Discount (lambda)
    double tdDiscount;   // Temporal difference discount
    double lastRew;      // Previous reward
    bool computeGrad;    // Should we learn by experience?

public:
   
    SARSAController(Approximator* approx, double tdDiscount);

    /**
     * Where the real coding is done to implement
     */
    virtual void getAction(Observation& obs, Vector& action, bool computeGrad = true);
    
    virtual void discountTrace();

    virtual void setDiscount(double discount);

    virtual void instantStep(Vector& reward);


};
}
#endif
