#ifndef CRFController_hh
#define CRFController_hh

// $Id$
class GridSim;

/**
 * The Conditional Random Field (CRF) controller which is NOT 
 * built upon any approximator.
 */

class CRFController : public Controller {


public:
   
    /**
	 *	variables
	 */    

    
    unsigned int numAgent;			// number of agents in the whole network

    /**
	 * Number of observations/inputs at each agent, assuming it is equal for all agents.
	 * E.g., there are 2 observations at each intersection: horizontal & vertical
	 * Each observation can be continuous real valued.
	 */ 
	unsigned int numObs;				

	/** 
	 * Number of action at each node (assuming each node has same # of action)
	 * Actions must be discrete and finite.
	 */
	unsigned int numAction;		
	
    /**
     * Eligibility traces, one for edge features and one for node features;
     */
 	/************************************************************************/
	/*  Notice: trace is served by node_fea_grad and edge_fea_grad.			*/
	/************************************************************************/
//	Matrix edgeTrace;	// trace of edges
//	Matrix nodeTrace;	// trace of nodes
    
    /**
     * Discount factor
     */
    double discount;
    
    /**
     * Step size
     */
    double stepSize;

    /**
     * the input command string for mcmc_tree sampler
     */
	stringstream input_string;

	/**
	 * The timer
	 */
	unsigned int time_length;

	/**
	 * The maximum number of samples drawn
	 */
	unsigned int max_num_sample;

	 
    /**
     * CRF only concerns about the global reward
     */
    // bool localRewards;

	void run_inference();

     	   
public:
    /**
	 *	Methods
	 */    
    
	
	CRFController(int size_x, int size_y, int obsRow, int actDim, int numEdgeFea, 
							 int numAgent, int numObs, int numAction, int argc, char** argv);
    virtual ~CRFController() {};	// destructor

    /**
     *  Where the real coding is done to implement 
     */
    virtual void getAction(Observation& obs, Vector& action, bool computeGrad = true);
    virtual void getMostProbAction(Observation& obs, Vector& action);

    virtual int	 getInputDim();
    virtual int	 getNumParams();
    virtual void discountTrace();
    virtual void setDiscount(double discount);
    virtual void accumulateGrad(Vector& reward, Observation& newObs);
    virtual void resetGrad();
    virtual void instantStep(Vector& reward);
    virtual void batchStep();
    virtual void computeDirection(int steps);
    virtual void setStepSize(double stepSize);
    virtual void resetTrace();
    virtual void resetParams();
    virtual void randomizeParams(double maxRand);
    virtual double getMaxParam();
    virtual void write(std::ostream& o);
    virtual void read(std::istream& o);

    virtual void reduce(Vector& v, Approximator::StatsEnum s);
    virtual void scatter(Vector& v, Approximator::StatsEnum s);
    
    /**
     * Specific to CRF
     */
	void sampleActionAndMean();	// read back the result of PGM
	void updateNodePotential();	// update the node potential in the PGM
	void initializeNetwork(int size_x, int size_y, int obsRow, int actDim, 
			int numEdgeFea, int argc, char** argv);	// initialize the network topology and potentials
private:
	void fill_trace();
};
#endif
