#include"PGBasics.hh"
#include "definitions.hh"
#include "Ps.hh"

#include"ScatsController.hh"
#include "NeuralNet.hh"
#include "BasicController.hh"
//#include "BinaryController.hh"
//#include "CyclicPolicyBias.hh"
#include "MixedController.hh"

using namespace libpg;

#define READING_START 7
#define READING_OFFSET 8


MixedController::MixedController(NeuralNet* nn, Controller* ac, Controller* tc,
				 int train_time) 
  : approximator(nn), trainer_controller(tc), approx_controller(ac), 
    training_time(train_time) {

  //approx_controller = new BinaryController(approximator);
  for(int i = 0; i < PHASES; i++) {
    statistics[i] = 0;
  }
  time = -1; // used to count the number of training iterations for the approximator
  trainer_dim = trainer_controller->getInputDim();
  approx_dim =  approx_controller->getInputDim();
  inputs = trainer_dim + approx_dim;
  reward.resize(1);
}


void MixedController::get_action_from_subcontroller
(Controller* controller, Observation& obs, Vector& action, 
 int split_from, int split_to, bool learn) {
  int dim1 = controller->getInputDim();
  assert(dim1 == split_to - split_from); 
  int dim2 = obs.getFeatures().size2();
  Observation sub_obs(dim1, dim2, dim2);
  sub_obs.getFeatures().assign(project(obs.getFeatures(), range(split_from, split_to), 
				  range(0, dim2)));
  sub_obs.setAgent(obs.getAgent());
  controller->getAction(sub_obs, action, learn); 
}

void MixedController::train_approximator(Observation& obs, Vector& action) {
  double correct_action = action[0];
  Vector output(action.size());
  int dim2 = obs.getFeatures().size2();
  Observation sub_obs(approx_dim, dim2, dim2);
  assert(approx_dim == inputs - trainer_dim);
  sub_obs.getFeatures().assign(project(obs.getFeatures(), range(trainer_dim, inputs), range(0, dim2)));
  sub_obs.setAgent(obs.getAgent());
  approx_controller->getAction(sub_obs, output, true);
  if((int)output[0] == correct_action) {
    reward(0) = 1;
  }
  else {
    reward(0) = 0;
  }
  //cout << "approximated action is " << output[0] << endl;
  //cout << "correct action is " << correct_action << endl;
  /*
  //cout << "approximated action is " << output[0] << endl;
  approximator->doApprox(sub_obs, output);
  double target = Ps::pNN_TARGET;
  if(correct_action == 1) {
  target = -target;
  }
  double delta = target - output[0];
  output[0] = delta;  
  approximator->resetTrace();
  approximator->feedbackGrad(sub_obs, output);
  //approximator->instantStep(1.0); // equivalent to reward of 1.0
  */
}

void MixedController::getAction(Observation& obs, Vector& action,  
				bool computeGrad) {   
  
  assert(action.size() == 1);
  //cout << approximator->getStepSize() << endl;
  //exit(1);
  
  // overwrite step size set by STSOptim 
  //(NOTE: this cannot be done in constructor unless we change setStepSize() and
  // setDiscount() to do nothing)
  if (time == -1) {
    approx_controller->setStepSize(Ps::pPRE_LEARNING_STEPSIZE);
    approx_controller->setDiscount(0);
    time++;
  }
  
  if(time == training_time ) {
    approx_controller->setStepSize(Ps::pSTEP_SIZE);
    approx_controller->setDiscount(Ps::pDISCOUNT);
    time++;
  } 

  //Vector o = column(obs.getFeatures(), obs.agent);
  /*
  cout << "Controller " << obs.agent << endl;
   for(int i = 0; i < inputs; i++) {
	cout << i <<":" << o(i) << " ";
      }
   cout << endl << endl;
  */

  if(time < training_time) { 
      get_action_from_subcontroller(trainer_controller, obs, action, 0,  
				    trainer_controller->getInputDim(), 
				    computeGrad); 
      if(computeGrad) {
	train_approximator(obs, action); 
	time++;
      }
    }
    else {
      get_action_from_subcontroller(approx_controller, obs, action, 
				    trainer_controller->getInputDim(), inputs, 
				    computeGrad);     
    } 

  //time++;
    
  /*
    //cout << "--------------------------" << endl;
    //action[0] = 1; // switch    
      for(int i = 0; i < inputs; i++) {
	cout << i <<":" << o(i) << " ";
      }     
      for(int i = 0; i < PHASES; i++) { 
	statistics[i] += phase_length[i]; 
	total_phases += phase_length[i]; 
      }     
      cout << "Time "<< time << " statistics: ";
      for(int i = 0; i < PHASES; i++) {
      cout << i << ": " << (statistics[i] / total_phases) << ",  ";
      }
      cout << endl;
      */
    //}
}

// not needed at the moment, return something
int MixedController::getInputDim() {
    return inputs;
}


void MixedController::write(ostream& o) { 
  // TODO: call NeuralNetworks write method
  approx_controller->write(o);
};  

void MixedController::read(istream& o) {
  //TODO: test
  /*
  cout << "reading settings" << endl;
  for(int i = 0; i < PHASES; i++) {
    o >> phase_length[i];
  }
  */
  approx_controller->read(o);
};

void MixedController::instantStep(Vector& external_reward) {
  if(time < training_time) {
    //cout << "backpropagating reward of " << reward << endl;
    approx_controller->instantStep(reward);
  }
  else {
    approx_controller->instantStep(external_reward);
  }
}
