#include"PGBasics.hh"
#include "definitions.hh"
#include "Ps.hh"
#include "intersection.hh"

#include"ScatsController.hh"

using namespace libpg;

#define READING_START 5
#define READING_OFFSET 8
#define READING_START2 21
//#define MAX_CYCLE_LENGTH 16
//#define MAX_CYCLE_LENGTH 4
//#define MAX_CYCLE_LENGTH 32
//#define MAX_PHASE_DURATION 13

//#define VERBOSE_SCATS true

ScatsController::ScatsController(int no_inputs, int init_phase_length,
				 double upper_threshold, 
				 double lower_threshold,
				 bool sudden_changes, bool RP) 
  : inputs(no_inputs), high_threshold(upper_threshold),  
    low_threshold(lower_threshold), boost(sudden_changes), random_phases(RP) {

  if(RP) {
    _MAX_CYCLE_LENGTH = Ps::pRPC_CYCLE_LENGTH;
    if(Ps::pMAX_CYCLE_LENGTH != Ps::pRPC_CYCLE_LENGTH) {
      cout << "Note: For comparison to usual Scats MAX_CYCLE_LENGTH "
	   << "must equal RPC_CYCLE_LENGTH!" << endl;
    }
  }
  else {
    _MAX_CYCLE_LENGTH = Ps::pMAX_CYCLE_LENGTH;
  }

  for(int i = 0; i < PHASES; i++) {
    phase_length[i] = init_phase_length;
    statistics[i] = 0;
  }
  //current_phase = 0;
  //duration = 0;
  //time = 0;
  total_phases = 0;
}

void ScatsController::getAction(Observation& obs, Vector& action,  
				bool computeGrad) {   

  //bool VERBOSE_SCATS = obs.agent == 1;
  bool VERBOSE_SCATS = false;
  double GOAL_SATURATION = Ps::pSCATS_GOAL_SATURATION;

  assert(action.size() == 1);

  Vector o;
  o.assign(column(obs.getFeatures(), obs.getAgent()));
  int current_phase = 0;
  int duration = 0;
  for(int i = 0; i < PHASES; i++) {
    if(o(i) == 1) {
      current_phase = i;
      break;
    }
  }
  duration = (int) o(PHASES);
  if(duration < phase_length[current_phase] - 1) {
    // no switch 
    if(!random_phases) {
    action[0] = 0; 
    }
    else {
      action[0] = current_phase;
    }
  }
  else {
    // switch
    int next_phase = (current_phase + 1) % PHASES;
    if(!random_phases) {
      action[0] = 1; 
    } 
    else {
      action[0] = next_phase;
    }
    if(next_phase == 0) {// full cycle completed, calculate new policy
      if(!Ps::pLEARN || !computeGrad) {
	if(VERBOSE_SCATS) {
	  cout << "not learning" << endl;
	}
	return;
      }

      if(VERBOSE_SCATS) {
	cout << endl;
	cout << "in Scats Controller " << obs.getAgent() << endl;
	for(int i = 0; i < inputs; i++) {
	  cout << i <<":" << o(i) << " "; 
	}
	cout << endl;
	cout << "phase length was: ";
	for(int i = 0; i < PHASES; i++) {
	  cout << i << ": " << phase_length[i] << ",  ";
	}
	cout << endl;
	cout << "current phase " << current_phase 
	     << ", duration " << duration << endl;
	cout << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! change !!!!!!!!!!!!!!!!!!!!!!!!!!!!!! " << endl;
      }
      int desired_phase_length[PHASES];
      double detec_counts[DIRECTIONS][QUEUES];
      //double summed_detec_counts[PHASES];
      double max_detec_counts[PHASES];
      double max_throughput[PHASES];
      //bool detec_active[DIRECTIONS][QUEUES];
      for(int d = 0; d < DIRECTIONS; d++) { 
	for(int q = 0; q < QUEUES; q++){
	  //int pos1 = READING_START + d * QUEUES + q;
	  //detec_active[d][q] = (bool) o(pos1);	  
	  int pos2 = READING_START + READING_OFFSET +
	    d * QUEUES + q;
	  detec_counts[d][q] = o(pos2);
	  if(VERBOSE_SCATS) {
	    cout << pos2 << ":" << detec_counts[d][q] << " ";
	  }
	}
      } 
      if(VERBOSE_SCATS) {
	cout << endl;
	cout << "max throughput in phase ";
      }
      for(int j = 0; j < PHASES; j++) {
	max_throughput[j] = o(READING_START2 + j);
	if(VERBOSE_SCATS) {
	  cout << j << ": " << max_throughput[j] << "  "; 
	}
      }
      if(VERBOSE_SCATS) {
	cout << endl;
      }
      /*
   summed_detec_counts[Intersection::NS_straight] = 
	detec_counts[Intersection::North][STRAIGHT_QUEUE] + 
	detec_counts[Intersection::South][STRAIGHT_QUEUE]; 
      summed_detec_counts[Intersection::NS_right] = 
	detec_counts[Intersection::North][RIGHT_QUEUE] + 
	detec_counts[Intersection::South][RIGHT_QUEUE]; 
      summed_detec_counts[Intersection::EW_straight] = 
	detec_counts[Intersection::East][STRAIGHT_QUEUE] + 
	detec_counts[Intersection::West][STRAIGHT_QUEUE]; 
      summed_detec_counts[Intersection::EW_right] = 
	detec_counts[Intersection::East][RIGHT_QUEUE] + 
	detec_counts[Intersection::West][RIGHT_QUEUE]; 
      */
      max_detec_counts[Intersection::NS_straight] = 
	max(detec_counts[Intersection::North][STRAIGHT_QUEUE], 
	    detec_counts[Intersection::South][STRAIGHT_QUEUE]); 
      max_detec_counts[Intersection::NS_right] = 
	max(detec_counts[Intersection::North][RIGHT_QUEUE],
	    detec_counts[Intersection::South][RIGHT_QUEUE]); 
      max_detec_counts[Intersection::EW_straight] = 
	max(detec_counts[Intersection::East][STRAIGHT_QUEUE], 
	    detec_counts[Intersection::West][STRAIGHT_QUEUE]); 
      max_detec_counts[Intersection::EW_right] = 
	max(detec_counts[Intersection::East][RIGHT_QUEUE],  
	    detec_counts[Intersection::West][RIGHT_QUEUE]); 

      if(VERBOSE_SCATS) {
	cout << "max detector counts: ";
	for(int j = 0; j < PHASES; j++) {   
	  cout << j << ": " << max_detec_counts[j] << "  ";
	}
	cout << endl;
      }

      double saturation[PHASES];
      if(VERBOSE_SCATS) {
	cout << "saturation: ";
      }
      for(int j = 0; j < PHASES; j++) {
	saturation[j] = max_detec_counts[j] / max_throughput[j];
	if(VERBOSE_SCATS) {
	  cout << j << ": " << saturation[j] << "  ";
	}
      }
      if(VERBOSE_SCATS) {
	cout << endl;
      }
      int cycle_length = phase_length[0] + phase_length[1] + phase_length[2] +
	phase_length[3];
      if(VERBOSE_SCATS) {
	cout << "cycle_length: "<< cycle_length << endl;
      }
      bool general_saturation = false;
      bool general_emptyness = false;

      if(Ps::pSCATS_CYCLE_EXTENSION) {
	if(saturation[0] > GOAL_SATURATION && 
	   saturation[1] > GOAL_SATURATION && 
	   saturation[2] > GOAL_SATURATION && 
	   saturation[3] > GOAL_SATURATION) {
	  general_saturation = true;
	if(VERBOSE_SCATS) {
	  cout << "general saturation" << endl;
	}
	}
	else if(saturation[0] < 0.2 && saturation[1] < 0.2 && saturation[2] < 0.2 && 
		saturation[3] < 0.2) {
	  general_emptyness = true;
	if(VERBOSE_SCATS) {
	  cout << "general emptyness" << endl;
	}
	}
      }
      if(general_saturation &&
	 cycle_length < (_MAX_CYCLE_LENGTH - PHASES)) {
	//cycle_length += PHASES;
	for(int j = 0; j < PHASES; j++) {
	  desired_phase_length[j] = min(phase_length[j] + 1, 
					Ps::pMAX_PHASE_DURATION);
	}
      }   
      else if(general_emptyness && 
	 cycle_length > PHASES) {
	//cycle_length -= PHASES;
	for(int j = 0; j < PHASES; j++) {
	  desired_phase_length[j] = max(1, phase_length[j] - 1);
	}
      }
      else {
	double total_saturation = saturation[0] + saturation[1] +
	  saturation[2] +saturation[3];
	if(VERBOSE_SCATS) {
	  cout << "total saturation: " << total_saturation << endl;
	}

	/* Note: the following only makes sense if there is no 
	   general emptyness
	if(total_saturation == 0) {
	  if(VERBOSE_SCATS) {
	    cout << "no change in policy"<< endl;
	  }
	  return; // no change in policy
	}
	*/

	for(int j = 0; j < PHASES; j++) {
	  if(saturation[j] == 0) {
	    desired_phase_length[j] = 1; // minimum: phase must come on for 1 interval
	    //cycle_length--;
	  }
	  else {
	    //int possible_throughput = 1 + (phase_length[j] - 1) * 5;
	    if(max_detec_counts[j] /  max_throughput[j] > 
	       GOAL_SATURATION) {
	      desired_phase_length[j] = phase_length[j] + 1;
	    }
	    else {
	    desired_phase_length[j] = 
	      max(1, (int) ceil((max_detec_counts[j] / 
	      			 GOAL_SATURATION - THROUGHPUT_START_UP) / 
				THROUGHPUT_FLOW + 1));
	    if(VERBOSE_SCATS) {
	      cout << "desired phase length [" << j << "]: " 
		   << desired_phase_length[j] << endl;
	    }
	    }
	    if(desired_phase_length[j] > Ps::pMAX_PHASE_DURATION) {
	      if(VERBOSE_SCATS) {
		cout << "desired phase length [" << j << "]: " 
		     << desired_phase_length[j] 
		     << " is too long, using MAX_PHASE_DURATION "
		     << Ps::pMAX_PHASE_DURATION << endl;
	      }
	      desired_phase_length[j] = Ps::pMAX_PHASE_DURATION;
	    }
	  }
	  /*
	    }
	    cout << "cycle_length: "<< cycle_length << endl;
	    cout << "relative saturations: ";
	    for(int j = 0; j < PHASES; j++) {
	    cout << j << ": " << saturation[j]  / total_saturation<< "  ";
	    }
	    for(int j = 0; j < PHASES; j++) {
	    if(saturation[j] != 0) {
	    phase_length[j] =   
	    min((int) (saturation[j] / total_saturation * cycle_length 
	    / PHASES), 
	    Ps::pMAX_PHASE_DURATION); 
	    cout << "saturation[j] / total saturation * cycle_length: "  
	    << saturation[j] / total_saturation * cycle_length
	    /  PHASES << endl;
	    
	    }
	    }
	  */
	}
      }
      bool desired_change = desired_phase_length[0] != phase_length[0]  || 
	desired_phase_length[1] != phase_length[1]  || 
	desired_phase_length[2] != phase_length[2]  || 
	desired_phase_length[3] != phase_length[3] ;

      if(VERBOSE_SCATS && desired_change) {  
      //if(VERBOSE_SCATS) {  
	cout << "desired change: "<< desired_change << 
	  ",  desired new policy: ";
	for(int i = 0; i < PHASES; i++) {
	  cout << i << ": " << desired_phase_length[i] << ",  ";
	}
	cout << endl;
      }

      bool change = desired_phase_length[0] != phase_length[0]  || 
	desired_phase_length[1] != phase_length[1]  || 
	desired_phase_length[2] != phase_length[2]  || 
	desired_phase_length[3] != phase_length[3] ;

      if(boost) {
	for(int i = 0; i < PHASES; i++) {
	  phase_length[i] = desired_phase_length[i];
	}
      }
      else {
	for(int i = 0; i < PHASES; i++) {
	  if(phase_length[i] < desired_phase_length[i]) {
	    phase_length[i]++;
	  }
	  else if(phase_length[i] > desired_phase_length[i]) {
	    phase_length[i]--;
	  }
	}
      }
       cycle_length = 
	phase_length[0] + phase_length[1] + 
	phase_length[2] + phase_length[3];
      int counter = 0;
      while(cycle_length > _MAX_CYCLE_LENGTH) {
	if(VERBOSE_SCATS) {
	  cout << "cycle length > _MAX_CYCLE_LENGTH" << endl;
	}
	if(phase_length[counter] > 1) {
	  phase_length[counter] --;
	  cycle_length--;
	}
	counter = ((counter + 1) % PHASES);
      }
      
      if(VERBOSE_SCATS && !change) {
	cout << "no change in policy" << endl;
      }
      else if(VERBOSE_SCATS && change) {  
	cout << "Policy now: ";
	for(int i = 0; i < PHASES; i++) {
	cout << i << ": " << phase_length[i] << ",  ";
	}
	cout << endl;
      }
    
      cycle_length = phase_length[0] + phase_length[1] + phase_length[2] +
	phase_length[3];
      assert(cycle_length <= _MAX_CYCLE_LENGTH);
      for(int i = 0; i < PHASES; i++) {
	assert(phase_length[i] <= Ps::pMAX_PHASE_DURATION);
      }
    }
      /*
      if(general_emptyness || general_saturation) {
	cout << "Policy now: ";
	for(int i = 0; i < PHASES; i++) {
	  cout << i << ": " << phase_length[i] << ",  ";
	}
      }
      cout << endl;
      
      cout << "Saturation now: ";
      for(int i = 0; i < PHASES; i++) {
	cout << i << ": " <<  summed_detec_counts[i] / 
	  (2+ (phase_length[i] - 1) * 10) << ",  ";
      }
      cout << endl;
      */
      /*
      double total_count = summed_detec_counts[0] + 
	summed_detec_counts[1] +
	summed_detec_counts[2] + 
	summed_detec_counts[3];
      
      if(total_count > 0) {
	for(int i = 0; i < PHASES; i++) {
	  //cout << summed_detec_counts[i] << " -> ";
	  summed_detec_counts[i] = summed_detec_counts[i] / total_count;
	  //cout << summed_detec_counts[i] << ",  ";
	}
	//cout << endl;
      }
      else{
	for(int i = 0; i < PHASES; i++) {
	  cout << summed_detec_counts[i] << ",  ";
	}
	cout << endl;
	}
      //double high_threshold = 0.5;
      //double low_threshold = 0.1;
    
      bool change = false;
      for(int i = 0; i < PHASES; i++) {
	if( (summed_detec_counts[i] > high_threshold) &&
	    (phase_length[i] < Ps::pMAX_PHASE_DURATION) ) {
	  //cout << "prolonging phase " << i << endl;
	  phase_length[i] = phase_length[i] + 1;

	  change = true;
	}
	else if( (summed_detec_counts[i] < low_threshold) &&
		 (phase_length[i] > 1) )  {
	  //cout << "shortening phase " << i << endl;
	  phase_length[i] = phase_length[i] - 1;
	  change = true;
	}
      }
     
      for(int i = 0; i < PHASES; i++) {
	statistics[i] += phase_length[i];
	total_phases += phase_length[i];
      }
    
      cout << "Time "<< time << " statistics: ";
      for(int i = 0; i < PHASES; i++) {
      cout << i << ": " << (statistics[i] / total_phases) << ",  ";
      }
      cout << endl;
      */
  }
}

int ScatsController::getPolicy(int phase) {
  return phase_length[phase];
}

// not needed at the moment, return something
int ScatsController::getInputDim() {
    return inputs;
}

void ScatsController::resetParams() { 
}

void ScatsController::randomizeParams() { 
  NYI; 
}

void ScatsController::write(ostream& o) { 
  for(int i = 0; i < PHASES; i++) {
    o << phase_length[i] << endl;
  }
};  

void ScatsController::read(istream& o) {
  
  //cout << "reading settings" << endl;
  for(int i = 0; i < PHASES; i++) {
    o >> phase_length[i];
    //cout << "read phase_length["<< i<< "]: "<< phase_length[i] << endl;
  }
};

