#ifndef SensorSim_hh
#define SensorSim_hh
// $Id$

/**
 * Container for any RL problem we care to define. We simply have to
 * define the dimensionality of  actions, a reward function, a way to get
 * observations, and a way to do actions.
 */

using namespace libpg;

class SensorSim : public Simulator {

private:

    int num_target;			// number of targets (at the outset)
	int num_cell;			// number of target cells (number of agents / 2 - 1)
	int num_sensor;			// number of sensor (num_hole * 2 + 2)
	int num_living_target;	// number of living targets
    int nextReward;		
	int num_agent;			// number of agent (num_sensor for NN distributed RL, num_hole + 1 for CRF)
	int *energy;			// the energy of all targets
	int *position;			// A target's current position
	int *buffer1, *buffer2;	// initialization working buffer.  Avoid repeated space allocation.
    bool *occupied;			// whether a position is occupied by a target
	int ctl_id;				// which type of communication, 
							// which determines format of observation to return, and action to take
	int cur_cycle;		// How many rounds that the simulator has taken so far in this epoch
	int max_cycle;			// How many steps each epoch is allowed to take in each epoch
	int total_killed;		// In each epoch, how many are killed
	int total_cycle;		// In each epoch, how many steps are taken
	int cur_epoch;
	int num_epoch_per_review;
	bool **store_occupied;
	int **store_position;
public:

    SensorSim(int *argc, char **argv);

    virtual ~SensorSim();

    /**
     * Get the number of rows in observation matrix. Generally the
     * length of the observation vector.
     */
    virtual int getObsRows();

    /**
     * Get the number of columns in the observation matrix Typically
     * use multiple columns to provide a different observation to each
     * agent, so the number of columns is the same as getAgents()
     */
    virtual int getObsCols();

    /**
     * Get the total number of agents. Usually 1.
     */
    virtual int getAgents();

    /**
     * Get the total action dimensionality. Often, but not
     * necessarily, the same as the number of agents.
     */
    virtual int getActionDim();

    /**
     * Get the dimensionality of the reward vector. Often, but not
     * necessarily, the same as the number of agents.
     */
    virtual int getRewardDim();

	/**
     * Just return the reward associated with the most recent state
     * achieved.  
     * @param rewards vector of rewards. The vector has
     * getAgents() dimensionality In non-local reward mode this vector
     * will just be averaged.
     */
    virtual void getReward(Vector& rewards);

    virtual void getObservation(Observation& obs);


    /**
     * Do action issued by a controller @return  vector of actions,
     * typically just a scalar which is cast to an int, but it there
     * are multiple agents it will be a vector of actions for each
     * agent.
     * @param action vector that describes the action to perform.
     */
    virtual int doAction(Vector& action);

	virtual double get_max_reward();

	void InitializeGame();

	void restartGame();

	virtual char * get_diagnosis()  {return diagnosis_msg;};
};


#endif
