#ifdef HAVE_QT
#include <qapplication.h>
#include <qevent.h>
#include <qwidget.h>
#include <qpainter.h>

#include <vector>
#include <list>
#include <iostream>

#include "PGBasics.hh"

#include "definitions.hh"
#include "Ps.hh"
#include "intersection.hh"
#include "car.hh"
#include "sts.hh"
#include "simgui.hh"

using namespace libpg;

#define SCALING 4

#define BORDER 10
#define IC_SCALE 50 // scales up intersection coordinates
#define LANE_WIDTH 6 // intersections are double the width of this, lights half
#define CAR_WIDTH 1
#define LIGHT_WIDTH 1

int OFFSET = SCALING * LANE_WIDTH; // needed such that intersection (0,0) 
//can be displayed fully

// (IC_SCALE - 2 * LANE_WIDTH) >= 0 must be true!

struct CarPosition {
  Intersection* last;
  Intersection* next;
  int pos_on_road;
  Intersection::Turn turn;

  CarPosition(Intersection* l, Intersection* n, int p, Intersection::Turn t) : 
    last(l), next(n), pos_on_road(p), turn(t){}

  bool operator<(const CarPosition& c2) const {
    bool smaller_turn = (turn < c2.turn);
    if((turn == Intersection::Straight || turn == Intersection::Left) &&
       (c2.turn == Intersection::Straight || c2.turn == Intersection::Left))
       smaller_turn = false; // Left and Straight turn queue are identical
    return last < c2.last || (last == c2.last && next < c2.next) || 
      (last == c2.last && next == c2.next && pos_on_road < c2.pos_on_road) ||
      (last == c2.last && next == c2.next && pos_on_road == c2.pos_on_road && 
       smaller_turn) ;
  }
};

 
WidgetTest::WidgetTest(STS& sim, QWidget *parent) : 
    QWidget(parent), simulator(sim) {
  cout << "creating Widget" << endl;
  int rows  = simulator.get_rows();
  int cols = simulator.get_cols();

  resize(SCALING * (IC_SCALE * (cols - 1) + 2 * LANE_WIDTH) + 2 * BORDER, 
	 SCALING * (IC_SCALE * (rows - 1) + 2 * LANE_WIDTH) + 2 * BORDER);
}

void drawRect(int x, int y, int w, int h, QPainter& painter) {
  painter.drawRect(SCALING * x + OFFSET + BORDER, 
		   SCALING * y + OFFSET + BORDER, 
		   SCALING * w, SCALING * h);
}
void drawLine(int x, int y, int x2, int y2, QPainter& painter) {
  painter.drawLine(SCALING * x + OFFSET + BORDER, 
		   SCALING * y + OFFSET + BORDER, 
		   SCALING * x2 + OFFSET + BORDER, 
		   SCALING * y2 + OFFSET + BORDER);
}

void drawIntersection(Intersection* i, QPainter& painter) {
  int x = i->get_x_coordinate();
  int y = i->get_y_coordinate();
  drawRect(x * IC_SCALE - LANE_WIDTH, 
		   y * IC_SCALE - LANE_WIDTH,
	   2 * LANE_WIDTH,  2 * LANE_WIDTH, painter);

  double phase_percent[PHASES];
  for(int j = 0; j < PHASES; j++) {
    phase_percent[j] = i->get_avg_phase_percent((Intersection::Phase) j);
    //cout << phase_percent[j] << " ";
    drawLine(x * IC_SCALE + 2 * (j - 2), 
	     (int) (y * IC_SCALE - phase_percent[j] * LANE_WIDTH),
	     x * IC_SCALE + 2 * (j - 2), 
	     (int) (y * IC_SCALE + phase_percent[j] * LANE_WIDTH), painter);
    }
  Intersection::Phase phase = i->get_current_phase();  

  QBrush brush;
  
  // draw NS_straight lights
  if(phase == Intersection::NS_straight)
    brush = QBrush(Qt::green);
  else
    brush = QBrush(Qt::red);
  painter.setBrush(brush);
  // left light at bottom
  drawRect(x * IC_SCALE  - LANE_WIDTH,
		   y * IC_SCALE + LANE_WIDTH - LIGHT_WIDTH,
	   LANE_WIDTH / 2, LIGHT_WIDTH, painter);
  // right light at top
  drawRect(x * IC_SCALE + (LANE_WIDTH / 2),
		   y * IC_SCALE - LANE_WIDTH,
	   LANE_WIDTH / 2, LIGHT_WIDTH, painter);

  // draw NS_right lights
  if(phase == Intersection::NS_right)
   brush = QBrush(Qt::green);
  else
    brush = QBrush(Qt::red);
  painter.setBrush(brush);
  // right light at bottom
  drawRect(x * IC_SCALE - (LANE_WIDTH / 2),
		   y * IC_SCALE + LANE_WIDTH - LIGHT_WIDTH,
	   LANE_WIDTH / 2, LIGHT_WIDTH, painter);
  // left light at top
  drawRect(x * IC_SCALE,
		   y * IC_SCALE - LANE_WIDTH,
	   LANE_WIDTH / 2, LIGHT_WIDTH, painter);
  
  // draw EW_straight lights
  if(phase == Intersection::EW_straight)
   brush = QBrush(Qt::green);
  else
    brush = QBrush(Qt::red);
  painter.setBrush(brush);
  // upper light at left
  drawRect(x * IC_SCALE - LANE_WIDTH, 
	   y * IC_SCALE - LANE_WIDTH, 
	   LIGHT_WIDTH, LANE_WIDTH / 2, painter); 
  // lower light at right
  drawRect(x * IC_SCALE + LANE_WIDTH - LIGHT_WIDTH, 
	   y * IC_SCALE + LANE_WIDTH / 2, 
	   LIGHT_WIDTH, LANE_WIDTH / 2, painter);

// draw EW_right lights
  if(phase == Intersection::EW_right)
   brush = QBrush(Qt::green);
  else
    brush = QBrush(Qt::red);
  painter.setBrush(brush);
  // upper light at right
  drawRect(x * IC_SCALE + LANE_WIDTH - LIGHT_WIDTH,
		   y * IC_SCALE,
	   LIGHT_WIDTH, LANE_WIDTH / 2, painter);
 // lower light at left
  drawRect(x * IC_SCALE - LANE_WIDTH, 
		   y * IC_SCALE - LANE_WIDTH / 2, 
	   LIGHT_WIDTH, LANE_WIDTH / 2, painter);

  
  painter.setBrush(Qt::NoBrush);
}

void drawRoad(int from_x, int from_y, int to_x, int to_y, QPainter& painter) {
  bool inc_x = (to_x  > from_x);
  bool inc_y = (to_y  > from_y);
  bool dec_x = (to_x < from_x);
  bool dec_y = (to_y < from_y);
  int sx, sy, ex, ey;
  sx = from_x * IC_SCALE;
  sy = from_y * IC_SCALE;
  ex = to_x * IC_SCALE;
  ey = to_y * IC_SCALE;
  if(inc_x) {
    sx += (LANE_WIDTH);
    ex -= (LANE_WIDTH);
  }
  else if(dec_x) {
    sx -= (LANE_WIDTH);
    ex += (LANE_WIDTH);
  }
  else if(inc_y) {
    sy += (LANE_WIDTH);
    ey -= (LANE_WIDTH);
  }
  else if(dec_y) {
    sy -= (LANE_WIDTH);
    ey += (LANE_WIDTH);
  }
  drawLine(sx, sy, ex, ey, painter); // middle line
  if(inc_x || dec_x) { // horizontal road
    drawLine(sx, sy + LANE_WIDTH, ex, ey + LANE_WIDTH, painter);
    drawLine(sx, sy - LANE_WIDTH, ex, ey - LANE_WIDTH, painter);
  }
  else { // vertical road
    drawLine(sx + LANE_WIDTH, sy,  ex + LANE_WIDTH, ey, painter);
    drawLine(sx - LANE_WIDTH, sy,  ex - LANE_WIDTH, ey, painter);
  }
}

void drawCars(CarPosition cp, int no_cars, QPainter& painter) {
  bool inexact = false;
  int dist = cp.pos_on_road;
  double tmp1 = (1.0 * IC_SCALE - 2 * LANE_WIDTH);
  double tmp2 =  1.0 * (Ps::pROAD_LENGTH + 1);
  double true_segment_length = (tmp1 / tmp2);
  //cout << "true segment length is " << true_segment_length << endl;
  //int segment_length = (int)((1.0 * IC_SCALE - 2 * LANE_WIDTH) / 
  //		     1.0 * (Ps::pROAD_LENGTH + 1));
  int segment_length = (int) true_segment_length;
  //cout << "segment length is " << segment_length << endl;
  int car_length = CAR_WIDTH * no_cars;
  if(car_length > segment_length) {
    //cout << "Warning: car length shortened by segment length, display inexact." << endl;
    car_length = segment_length;
    inexact = true;
  }
  //cout << "car length is " << car_length << endl;

  int from_x = cp.last->get_x_coordinate();
  int from_y = cp.last->get_y_coordinate();
  int to_x = cp.next->get_x_coordinate();
  int to_y = cp.next->get_y_coordinate();
  bool inc_x = (to_x  > from_x);
  bool inc_y = (to_y  > from_y);
  bool dec_x = (to_x < from_x);
  bool dec_y = (to_y < from_y);
  int x, y, width, height;
  x = to_x * IC_SCALE;
  y = to_y * IC_SCALE;
  if(inc_x || dec_x) {
    width = car_length;
    height = CAR_WIDTH;
  }
  else {
    width = CAR_WIDTH;
    height = car_length;
  }
  if(dist == 0) { // car waiting at intersection to
    bool sl_turn = (cp.turn == Intersection::Straight) || (cp.turn == Intersection::Left);
    if(inc_y) {
      y -= LANE_WIDTH + car_length;
      if(sl_turn)
	x += LANE_WIDTH - CAR_WIDTH;
      else
	x += (LANE_WIDTH / 2) - CAR_WIDTH;
    }
    else if(dec_x) {
      x += LANE_WIDTH;
      if(sl_turn)
	y += LANE_WIDTH - CAR_WIDTH;
      else
	y += (LANE_WIDTH / 2) - CAR_WIDTH;
    }
    else if(dec_y) {
      y += LANE_WIDTH;
      x -= CAR_WIDTH;
      if(sl_turn)
	x -= LANE_WIDTH - CAR_WIDTH;
      else
	x -= (LANE_WIDTH / 2) - CAR_WIDTH;	
    }
    else if(inc_x) {
      x -= LANE_WIDTH + car_length;
      y -= CAR_WIDTH;
      if(sl_turn)
	y -= LANE_WIDTH - CAR_WIDTH;	
      else
	y -= (LANE_WIDTH / 2) - CAR_WIDTH;
    }
  }
  else { // car on road, use next intersection as starting point for position calculation
    if(inc_x) {
      //cout << "x was " << x;
      x -= LANE_WIDTH + dist * segment_length;
      y -= CAR_WIDTH;
      y -= (LANE_WIDTH / 2);
      x -= car_length; 
      //cout << ", x is now " << x << endl << endl;
    }
    else if(dec_x) {
      x += LANE_WIDTH + dist * segment_length;
      y += (LANE_WIDTH / 2);
    }
    else if(inc_y) {
      //cout << "y was " << y;
      y -= LANE_WIDTH + dist  * segment_length;
      y -= car_length;
      x += (LANE_WIDTH / 2);
      //cout << ", y is now " << y << endl << endl;
    }
    else if(dec_y) {
      y += LANE_WIDTH + dist * segment_length; 
      x -= CAR_WIDTH;
      x -= (LANE_WIDTH / 2);
      //cout << ", y is now " << y << endl << endl;
    }
  }
  QBrush brush (Qt::blue);
  if(inexact)
    brush  = QBrush(Qt::yellow);
  painter.setBrush(brush);
  drawRect(x, y, width, height, painter);
  painter.setBrush(Qt::NoBrush);
}

void WidgetTest::paintEvent(QPaintEvent *event) {
  		
    QPainter painter(this);

    IntersectionVec intersections = simulator.get_intersections();
    IntersectionVec::iterator i;
    for(i = intersections.begin(); i < intersections.end() ; i++) {
      drawIntersection(*i, painter);
      for(int j = 0; j < DIRECTIONS; j++) {
	Intersection::Direction dir = (Intersection::Direction) j;
	if((*i)->outgoing[dir] != NULL) {
	  //cout << "in GUI: ";
	  //(*i)->print_id();
	  //cout << "has neighbour in direction " << dir << endl; 
	  Intersection* neighbour = (*i)->outgoing[dir];
	  //cout << "neighbour is ";
	  //neighbour->print_id();
	  //cout << endl;
	  drawRoad((*i)->get_x_coordinate(), (*i)->get_y_coordinate(), 
		   neighbour->get_x_coordinate(), 
	  	   neighbour->get_y_coordinate(), painter);
	}
      }
    }

    CarList& cars = simulator.get_cars();
    CarList::iterator j;
    map<CarPosition, int> cars_at_position;
    map<CarPosition, int>::iterator h;
    for(j = cars.begin(); j != cars.end(); j++) {
      Car* car = *j;
      int dist = car->get_position_on_road();
      Intersection* last = 0;
      Intersection* next = 0; 
      Intersection::Turn turn = Intersection::Straight;
      if(dist > 0) {
	last = car->get_current_intersection();
	next = car->get_next_intersection();
      }
      else {
	last = car->get_last_intersection();
	next = car->get_current_intersection();
	turn = car->get_turn();
      }
      CarPosition cp = CarPosition(last, next, dist, turn);
      if(cars_at_position.find(cp) != cars_at_position.end()) {
	//cout << "found " << cars_at_position[cp] << 
	//" cars at position, adding" << endl;
	cars_at_position[cp] = cars_at_position[cp] + 1;
      }
      else {
	//cout << "inserting car into map" << endl;
	cars_at_position[cp] = 1;
      }
    }
    for(h = cars_at_position.begin(); h != cars_at_position.end(); h++) {
    drawCars((*h).first, (*h).second, painter);
    }  
}
  
#endif
