#include <iostream>
#include<iomanip>
#include <vector>
#include <list>
#include "time.h"

#include "PGBasics.hh"

#include "definitions.hh"
#include "Ps.hh"
#include "intersection.hh"
#include "car.hh"
#include "scenario.hh"
#include "sts.hh"

using namespace libpg;

#define BLOCKED_FACTOR 50 // how much punishment a blocked car results in

STS::STS(Scenario* sc) : scenario(sc) {
  grid_rows = scenario->get_rows();
  grid_cols = scenario->get_cols();
  scenario->make_intersections(intersections, grid_rows, grid_cols);
  common_init();
}

/*
STS::STS(CarList the_cars, IntersectionVec the_intersections):  
  cars(the_cars), intersections(the_intersections) { 
  grid_rows = grid_cols =  3; // default, must be set by user
  common_init(); 
} 

STS::STS(int rows, int cols) { 
  grid_rows = rows; 
  grid_cols = cols; 
  intersections = make_intersection_grid(grid_rows, grid_cols);
  common_init(); 
}
*/
void STS::reset() {
  cars.clear();
  sim_time = 0;
  next_id = cars.size();
  statistics.clear();
  IntersectionVec::iterator i;
  for(i = intersections.begin(); i < intersections.end(); i++) {
    (*i)->reset();
  }

  /*
  Vector first_action(getActionDim());
  for(int i = 0; i < getActionDim(); i ++) {
    first_action[i] = 0;
  }
  doAction(first_action);
  */
}

void STS::common_init() {
  
  sim_time = 0;
  next_id = cars.size();
  reward_last_time_step = 0;
  sum_traveltime_last_time_step = 0; 
  cars_failed_last_time_step = 0;
  cars_arrived_last_time_step = 0;
  no_controlled_intersections = 0;
  IntersectionVec::iterator i;
  for(i = intersections.begin(); i !=  intersections.end(); i++) {
   if(!((*i)->is_dummy())) {
      no_controlled_intersections++;
    }
 }
  //cout << "controlled inters: " << no_controlled_intersections << endl;
 
  statisticsNames.push_back("Avg. TP.");
  statisticsNames.push_back("#Cars");
  statisticsNames.push_back("Blocked");
  statisticsNames.push_back("Avg. Wait");
  statisticsNames.push_back("Traveltime"); 
  statisticsNames.push_back("Q. length");
  statistics.resize(statisticsNames.size() + 2); // Traveltime and Wait time
  //will each be calculated from 2 accumulated numbers
  statistics.clear();

  save_history_for_intersection = -1;
  save_history_from = -1;
  save_history_to = -1;
  intersec_summary_file_name = "tmp";
  history_file = NULL;
  stat_file = NULL;

  if(Ps::pVERBOSE || VERBOSE_STS)
    cout << "STS created." << endl;
  //print_map();

  // Stuff the simulator class insist on
  actionDim = intersections.size();
  obsRows = 0;
  int obs_learner = OBSERVATIONS_LEARNER_BASIC;
  if(Ps::pUSE_NEIGHBOUR_INFO) {
    obs_learner += OBSERVATIONS_LEARNER_NEIGHBOURS;
  }
  if(Ps::pUSE_DETEC_ACTIVE_INFO) {
    obs_learner += OBSERVATIONS_LEARNER_DETEC_ACTIVE;
  }
  if(Ps::pUSE_DETEC_HISTORY_INFO) {
    obs_learner += OBSERVATIONS_LEARNER_DETEC_HISTORY;
  }
  if(Ps::pUSE_PHASE_HISTORY_INFO) {
    obs_learner += OBSERVATIONS_LEARNER_PHASE_HISTORY;
  }
  if(Ps::pUSE_PHASE_DURATION_INFO) {
    obs_learner += OBSERVATIONS_LEARNER_PHASE_DURATIONS;
  }
  if(Ps::pUSE_CURRENT_PHASE_DURATION_INFO) {
    obs_learner += OBSERVATIONS_LEARNER_CURRENT_PHASE_DURATION;
  }
  if(Ps::pUSE_CURRENT_PHASE_INFO) {
    obs_learner += OBSERVATIONS_LEARNER_CURRENT_PHASE;
  }
  if(Ps::pUSE_CYCLE_DURATION) {
    obs_learner += Ps::pRPC_CYCLE_LENGTH;
  }

  switch(Ps::pCONTROLLER) {
  case _LearningController:
    obsRows = obs_learner;
    break;
  case _FixedPhaseController: 
    obsRows = OBSERVATIONS_SCATS;
    break;
  case _ScatsController:
    obsRows = OBSERVATIONS_SCATS;
    break;
  case _FixedController: 
    obsRows = OBSERVATIONS_SCATS;
    break;
  case _MixedController: 
    obsRows = OBSERVATIONS_SCATS + obs_learner;
    break;
  case _MixedRPController: 
    obsRows = OBSERVATIONS_SCATS + obs_learner;
    break;
  case _RandomPhaseController:
    obsRows = obs_learner;
    break;
  case _Scenario6Controller:
    obsRows = obs_learner;
    break;
  case _Scenario4Controller:
    obsRows = OBSERVATIONS_SCATS + obs_learner;
    break;
  case _CRFController:
    obsRows = obs_learner;
    break;
  default:
    obsRows = OBSERVATIONS_SCATS;
  }
  obsCols = intersections.size();
  actions = STS_ACTIONS_PER_INTERSECTION;
  agents = intersections.size();

}

void STS::print_map() {
  IntersectionVec::iterator i;
  for (i = intersections.begin(); i < intersections.end() ; i++) 
    (*i)->print_connections();
}

void STS::run() {
  /* This method runs the simulator for some time without an external controller.
   Phases have equal length and merely alternate in order. */

  if(Ps::pVERBOSE || VERBOSE_STS)
    cout << "STS started." << endl << endl;
  double avg_reward = 0;
  //int start_time = time(NULL);
  for(int i = 0; i < 1; i++) {
  //while(!cars.empty()) {
    //while(time(NULL) - start_time < 10) {
    Vector phase_change(intersections.size());
    for(unsigned int j = 0; j < intersections.size(); j++)
      phase_change[j] = ((((int)sim_time) % 2) == 0); // switch phase every 2 intervals
    doAction(phase_change);
    //cout << "reward at iteration " << sim_time << " is " << getReward() << endl;
    avg_reward += get_global_reward();
  }
  /*
  avg_reward = avg_reward / sim_time ;
  cout << "Iterations: " << sim_time << endl;
  cout << "Cars arrived: " << cars_arrived << endl;
  cout << "Avg. reward> " << avg_reward << endl << endl;
  */
}

void STS::save_history(int from, int to, string file) {
  save_history_from = from;
  save_history_to = to;
  intersec_summary_file_name = file;
  history_file = new ofstream("detailed_history");
  if (!history_file->is_open()) {
    cout<<"!!! Failed to open "<<endl;
  }
  else
    save_detailed_history = true;
}

void STS::write_tracked_intersections_statistics() {
  if(stat_file == NULL) {
    stat_file = new ofstream("statistics");
    //stat_file << setprecision(3);
    if (!stat_file->is_open()) {
      cout<<"!!! Failed to open "<<endl;
      return;
    }
  }
  (*stat_file) << "Time Step "<< sim_time << ":  " << endl;
  IntersectionVec::iterator i;
  for(i = tracked_intersections.begin(); i < tracked_intersections.end(); i++) {
    (*stat_file) << "Intersection (" << (*i)->get_x_coordinate() << "," 
		 << (*i)->get_y_coordinate() << "): cars passed ";
 for(int j = 0; j < PHASES; j++) {
   (*stat_file)  <<  (*i)->get_cars_passed((Intersection::Phase) j) << " "; 
    }
    //(*stat_file) << endl;
    (*stat_file) << ",  wait times ";
    for(int j = 0; j < PHASES; j++) {
      (*stat_file)  <<  (*i)->get_avg_wait_time((Intersection::Phase) j) << " "; 
    }
    (*stat_file) << ",  avg wait time " << (*i)->get_avg_wait_time();
    (*stat_file) << endl;

  (*stat_file) << ",  phase durations ";
    for(int j = 0; j < PHASES; j++) {
      (*stat_file)  <<  (*i)->get_avg_phase_duration((Intersection::Phase) j) << " "; 
    }
    (*stat_file) << endl;
    (*stat_file) << "phase probabilities: ";
    for(int s = 0; s < Ps::pRPC_CYCLE_LENGTH; s++) {
      if(s % (int) (Ps::pRPC_CYCLE_LENGTH / 4) ==  0) (*stat_file) << endl;
      (*stat_file) << "(";
      for(int p = 0; p < PHASES; p++) {
	(*stat_file) << (*i)->get_avg_phase_count(s, (Intersection::Phase)p) << " ";
      }
      (*stat_file) << ") ";
    }
    (*stat_file) << endl;
    (*i)->reset_statistics(); 
    (*stat_file) << endl;
  }
  (*stat_file) << endl << endl;
}

void STS::write_snapshot() {
  (*history_file) << "Time Step "<< sim_time << ":  " << endl;
  IntersectionVec::iterator i;
  for(i = tracked_intersections.begin(); i < tracked_intersections.end(); i++) {
    (*history_file) << "Intersection (" << (*i)->get_x_coordinate() << "," 
		    << (*i)->get_y_coordinate() << "): " 
		    <<  (*i)->get_current_phase() << endl;
    //(*history_file) << (*i)->get_current_phase() << "  ";
  }
}

int STS::doAction(Vector& action){
  /* Simulates the system for one time step: sets the phases of the 
     intersections according to the parameter action and updates the 
     car positions (including deleting and producing new ones.) */
  if(Ps::pVERBOSE || VERBOSE_STS) {
    cout << "Time step " << sim_time << endl;
  }
  if(sim_time == save_history_from) {
    cout << "beginning to save, time step " << sim_time << endl;
    IntersectionVec::iterator i;
    for(i = tracked_intersections.begin(); i < tracked_intersections.end(); i++) {
      ostringstream ostr;
      ostr << intersec_summary_file_name << "_intersection_" <<
	(*i)->get_x_coordinate() <<
	(*i)->get_y_coordinate();
      string fn = ostr.str();
      (*i)->set_history_saving(fn);
    }
  }
  else if(sim_time == save_history_to) {
    //cout << "stopping to save" << endl;
    IntersectionVec::iterator i;
    for(i = tracked_intersections.begin(); i < tracked_intersections.end(); i++) {
      (*i)->unset_history_saving();
    }
    //intersections.at(save_history_for_intersection)->unset_history_saving();
  }
  //cout << action << endl;
  //int last_cars_arrived_last_time_step = cars_arrived_last_time_step;

  cars_arrived_last_time_step = 0; // reward needs this
  cars_failed_last_time_step = 0; 
  sum_traveltime_last_time_step = 0;
  reward_last_time_step = 0;

  update_intersections(action);
  update_cars();
  scenario->produce_cars(this);

  if(sim_time >= save_history_from && sim_time <= save_history_to &&  
     save_detailed_history) {
    write_snapshot();
  }
  if(((int) sim_time) % 9216 == 0) { // should be dividable by 16
    write_tracked_intersections_statistics();
  }
  save_statistics();
  if(Ps::pVERBOSE || VERBOSE_STS)
    cout << "------------------------------------------" << endl;
  sim_time++;

  return 0;
}

void STS::update_intersections(Vector& phase_changes) {
  if(Ps::pVERBOSE || VERBOSE_STS)
    cout << "in update_intersections" << endl;
  IntersectionVec::iterator i;
  int position = 0;
  for (i = intersections.begin(); i < intersections.end() ; i++) {
      int c = (int)phase_changes[position];
      //if(Ps::pCONTROLLER == _RandomPhaseController ||
      // Ps::pCONTROLLER == _FixedController) {
      if(Ps::pCONTROLLER == _RandomPhaseController ||
		Ps::pCONTROLLER == _FixedRPController ||
		Ps::pCONTROLLER == _MixedRPController ||
		Ps::pCONTROLLER == _Scenario4Controller || 
		Ps::pCONTROLLER == _CRFController) 
	  { //Scenario4 is MixedRP
		(*i)->update((Intersection::Phase) c); 
      }
      else {
		(*i)->update((bool)c); 
      }
      position++;
  }
}

void STS::update_cars() {
  // case A: decrease road_position, if turns to zero: check arrival,
  // (remove car, increment reward statistics), else calculate turn, 
  // send car_arrive to intersection. If queue_pos zero, send car_depart, 
  // otherwise set queue_position in car
  // case B: road position is zero from beginning, queue_position > zero,
  // decrease queue_position (by 1 or 5)
  if(cars.empty())
    return;
  CarList::iterator i = cars.begin();
  do {
    if(Ps::pVERBOSE || VERBOSE_STS) {
      cout << "updating car " << (*i)->get_id() << endl;
    }
    Car* car = *i;
    bool arrived = car->update();
    if(arrived) {
      if(Ps::pVERBOSE || VERBOSE_STS) {
	cout << "car has arrived at destination" << endl;
      }
      // TODO: calculate reward statistics
      cars_arrived_last_time_step++;
      sum_traveltime_last_time_step += sim_time - car->get_create_time();
      if(Ps::pVERBOSE || VERBOSE_STS) {
	cout << "traveltime was " << sim_time - car->get_create_time() << endl;
      }
      //update_reward(car);
      //Car* car = *i;
      i = cars.erase(i);
      delete(car);
    }
    else // only advance pointer if previous car was not deleted
      i++;
  }
  while(i !=cars.end());
}

void STS::clearStatistics() {
  statistics.clear();
}

void STS::save_statistics() {
  IntersectionVec::iterator i;
  for(i = intersections.begin(); i !=  intersections.end(); i++) {
    if(!((*i)->is_dummy())) {
      statistics(0) += (*i)->get_throughput_last_step();
      statistics(3) += (*i)->get_wait_times_last_step(); // total wt of cars that passed
      statistics(6) += (*i)->get_cars_passed_last_step();
      statistics(5) += (*i)->get_waiting_cars_last_step();
    }
  }
  statistics(1) += cars.size();
  statistics(2) += cars_failed_last_time_step;
  statistics(4) +=  sum_traveltime_last_time_step; 
  statistics(7) += cars_arrived_last_time_step;
/*
  if (sum_traveltime_last_time_step)
  {
	  cout<<sum_traveltime_last_time_step<<endl;
	  if (cur_step % 100 == 0)
	  {
		  cin.get();
	  }
  }
*/
  
}

std::vector<string>& STS::getStatisticsNames() {
  return statisticsNames;
}

Vector& STS::getStatistics(int steps) {
  statistics(0) /= (steps * no_controlled_intersections);
  statistics(1) /= steps;
  statistics(2) /= steps;
  statistics(3) = statistics(3) / statistics(6) / no_controlled_intersections;
  //statistics(3) /= (no_controlled_intersections);
  statistics(4) = statistics(4) / statistics(7);
  statistics(5) /= (steps * no_controlled_intersections);
  return statistics;
}

double STS::get_global_reward(){
  //return reward_last_time_step;
  return  - (double) (cars.size() + BLOCKED_FACTOR * cars_failed_last_time_step);
}

void STS::getReward(Vector& reward) 
{
	if(!Ps::pLOCAL_REWARDS)
		reward[0] = get_global_reward();
	else 
	{
		int pos = 0;
		IntersectionVec::iterator i;
		for (i = intersections.begin(); i < intersections.end() ; i++) 
		{
			reward(pos) = 0;
			if(Ps::pQUEUE_LENGTH_REWARDS)
				reward(pos) -= ((*i)->get_waiting_cars_last_step() != 0);
			if(Ps::pTHROUGHPUT_REWARDS)
//				reward(pos) += ((*i)->get_throughput_last_step() > 0 ? 1 : 0);
				reward(pos) += (*i)->get_throughput_last_step();
				pos++;
		}
		if(pos != (signed int) reward.size()) 
		{
			cout << "error in STS::getReward()!" << endl;
			exit(1);
		}
	}
}

void STS::produce_car(int origin, int destination) {
  double id = next_id;
  if(next_id == (2 << 29)) {
    next_id = 0; 
  }
  else {
    next_id++;
  }
  Car* c = new Car(intersections.at(origin), 
		   intersections.at(destination), sim_time, id);
  if(Ps::pVERBOSE || VERBOSE_STS) {
    cout << "Produced car with origin ";
    intersections.at(origin)->print_id();
    cout << " and destination ";
    intersections.at(destination)->print_id();
    cout << "." << endl;
  }
  if(!c->get_next_intersection()->is_road_free(c->get_next_incoming_direction())) {
    if(Ps::pVERBOSE || VERBOSE_STS) {
      cout << "Could not enter car into system, road is full." << endl;
    }
    cars_failed_last_time_step++;   
    delete(c);
  }
  else {
    c->get_next_intersection()->
      enter_car_on_road(c->get_next_incoming_direction(), 
			c->get_position_on_road()); 
    cars.push_back(c);   
  }
}

void STS::getObservation(libpg::Observation& obs) {
  
  switch(Ps::pCONTROLLER) {
  case _LearningController:
    get_observation_for_learning_controller(obs);
    break;
  case _FixedPhaseController: 
    get_observation_for_scats_controller(obs);
    break;
  case _ScatsController:
    get_observation_for_scats_controller(obs);
    break;
  case _FixedController: 
    get_observation_for_scats_controller(obs);
    break;
  case _MixedController: 
    get_observation_for_mixed_controller(obs);
    break;
  case _MixedRPController: 
    get_observation_for_mixed_controller(obs);
    break;
  case _RandomPhaseController:
    get_observation_for_learning_controller(obs);
    break;
  case _Scenario6Controller:
    get_observation_for_learning_controller(obs); 
    break;
  case _Scenario4Controller:
    get_observation_for_mixed_controller(obs); 
    break;
  case _CRFController:
    get_observation_for_learning_controller(obs); 
    break;
  default:
    return get_observation_for_scats_controller(obs);
  }
}

void STS::get_observation_for_learning_controller(Observation& obs) 
{
	int col = 0;
	IntersectionVec::iterator i;
	for (i = intersections.begin(); i < intersections.end() ; i++) 
	{
		/*
		if(Ps::pVERBOSE || VERBOSE_STS) { 
		  cout << "Learner Observation for "; 
		  (*i)->print_id(); 
		  cout << ":" << endl; 
		}
		*/
		int obs_learner = obs.getFeatures().size1(); 
		//int obs_learner = obsRows;
		//if(Ps::pCONTROLLER == _MixedController) {
		// obs_learner -= OBSERVATIONS_SCATS;
		//}

		obs.getFeatures()(0, col) = 1; // 1 bit for constant 1 input
//		obs.getFeatures()(0, col) = 0; // 1 bit for constant 1 input

		for (int j = 1; j < obs_learner; j++)
		  obs.getFeatures()(j, col) = 0; // initialization

		int next_entry = 1;
		int phase = (*i)->get_current_phase();

		if (Ps::pUSE_CURRENT_PHASE_INFO) 
		{
			obs.getFeatures()(phase + 1, col) = 1; // 4 bits for current phase
			next_entry += PHASES;
		}
		if (Ps::pUSE_CURRENT_PHASE_DURATION_INFO) 
		{
			// 5 bits for duration of current phase, goes from 1 to 13
			int duration =  (*i)->get_phase_counts(phase); 
 
			if (duration >= 1)
				obs.getFeatures()(next_entry, col) = 1;
			next_entry++;
			if (duration >= 2)
				obs.getFeatures()(next_entry, col) = 1;
			next_entry++;
			if (duration >= 4)
				obs.getFeatures()(next_entry, col) = 1;
			next_entry++;
			if (duration >= 8)
				obs.getFeatures()(next_entry, col) = 1;
			next_entry++;
			if (duration >= 13) 
				obs.getFeatures()(next_entry, col) = 1;
			next_entry++;
		}

		if (Ps::pUSE_CYCLE_DURATION)
		{
			int cd = (*i)->get_cycle_duration();  // RPC_CYCLE_LENGTH bits
			// for how long the cycle has been on (especially necessary for RPC)
			cd = min(cd, Ps::pRPC_CYCLE_LENGTH - 1);
			if(cd == -1) 
				cd = Ps::pRPC_CYCLE_LENGTH - 1; // especially for restarts in offset scenario...
			obs.getFeatures()(next_entry+cd, col) = 1;
			next_entry += Ps::pRPC_CYCLE_LENGTH;
		}

		if (Ps::pUSE_PHASE_DURATION_INFO) 
		{
			for (int j = 0; j < PHASES; j++)
			{	
				// 4*5 bits for duration of other phases in last cycle
				int duration =  (*i)->get_phase_counts(j); 
				//cout << "duration is " << duration << endl;
				if (duration >= 1)
					obs.getFeatures()(next_entry, col) = 1;
				next_entry++;
				if (duration >= 2)
					obs.getFeatures()(next_entry, col) = 1;
				next_entry++;
				if (duration >= 4)
					obs.getFeatures()(next_entry, col) = 1;
				next_entry++;
				if (duration >= 8)
					obs.getFeatures()(next_entry, col) = 1;
				next_entry++;
				if (duration >= 13)
					obs.getFeatures()(next_entry, col) = 1;
				next_entry++;
			}
		}
		if(Ps::pUSE_PHASE_HISTORY_INFO) 
		{
			for(int j = 0; j < PHASES; j++) 
			{
				// 4 doubles for phase percentage history
				obs.getFeatures()(next_entry, col) = (*i)->get_avg_phase_percent(j);
				next_entry++;
			}
		}
		if(Ps::pUSE_DETEC_ACTIVE_INFO) 
		{  
			// 4 * 2 bits whether detector is active
			for(int d = 0; d < DIRECTIONS; d++) 			
				for(int q = 0; q < QUEUES; q++) 
				{
					if((*i)->is_detector_active(d, q)) 
					{
						obs.getFeatures()(next_entry, col) = 1;
						//cout << "detector["<< d <<"][" << q <<"] is active" << endl;
					}
					else
					{
						//cout << "detector["<< d <<"][" << q <<"] is not active" << endl;
					}
					next_entry++;
				}	
		}
		if(Ps::pUSE_DETEC_HISTORY_INFO) 
		{
			// 4 * 2 * 3 bits: past detector activity
			for(int d = 0; d < DIRECTIONS; d++) 
				for(int q = 0; q < QUEUES; q++)
				{
					int count = (*i)->get_detector_count_last_step(d, q);
					int max_cars = (*i)->get_car_throughput();
					//cout << "count is " << count << ", max_cars is " << max_cars << endl;
					if(count > 0)
						obs.getFeatures()(next_entry, col) = 1;
					next_entry++;
					if(count > max_cars / 2)
						obs.getFeatures()(next_entry, col) = 1;
					next_entry++;
					if(count == max_cars)
						obs.getFeatures()(next_entry, col) = 1;
					next_entry++;
				}
		}
		if(Ps::pUSE_NEIGHBOUR_INFO) 
		{
			// 4 * 4 bits for neighbours' phases
			for(int d = 0; d < DIRECTIONS; d++) 
			{ 
				if((*i)->incoming[d] != 0) 
				{
					int neighbour_phase = (*i)->get_neighbours_phase(d);
					obs.getFeatures()(next_entry + neighbour_phase, col) = 1;
				}
				next_entry += PHASES;
			}

			// changed to binary neighbour obs:        5 * 2 bits
			// first is EW traffic, second is NS traffic: each is 1 if it is bigger 

			// ugly: we assume a max. road length of 5 (and min of 2)
			int max_road_length = 5;
			int min_road_length = 2;
			//int max_road_length = Ps::pROAD_LENGTH;
			//int min_road_length = Ps::pROAD_LENGTH;
			for (int j = min_road_length; j <= max_road_length; j++) 
			{
				//int ew_ns_diff  =  (*i)->get_binary_neighbouring_detec_counts(j - 1);
				int ew_ns_diff  =  (*i)->get_binary_neighbouring_detec_counts(j-1);
				if(ew_ns_diff  < 0)
				{
				  //cout << "ew_traffic is bigger" << endl;
				  obs.getFeatures()(next_entry, col) = 1;
				}
				else if(ew_ns_diff  > 0){
				  //cout << "ns_traffic is bigger" << endl;
				  obs.getFeatures()(next_entry + 1, col) = 1;
				}
				next_entry +=2;
			}
			/*
			// 4 doubles for possible incoming traffic
			for(int d = 0; d < DIRECTIONS; d++){ 
			obs.getFeatures()(next_entry, col) =  
			 (*i)->get_neighbouring_detec_counts(d); 
			next_entry++; 
			}
			*/
		}
		//cout << "next entry " << next_entry << endl;
		assert(next_entry == obs_learner);

		/*
		if(Ps::pVERBOSE || VERBOSE_STS) { 
		  for(int j = 0; j < 32; j++) 
			cout << " " << obs.getFeatures()(j, col); 
		  cout << endl; 
		  for(int j = 32; j < OBSERVATIONS_LEARNER; j++) 
			cout << " " << obs.getFeatures()(j, col); 
		  cout << endl; 
		  } 
		*/
		col++;
	}
	//cout << endl;
}

void STS::get_observation_for_scats_controller(Observation& obs) {
  int col = 0;
  IntersectionVec::iterator i;
  for (i = intersections.begin(); i < intersections.end() ; i++) {
    /*
    if(Ps::pVERBOSE || VERBOSE_STS) { 
      cout << "SCATS Observation for "; 
      (*i)->print_id(); 
      cout << ":" << endl; 
      }
    */
    assert(obs.getFeatures().size1() == OBSERVATIONS_SCATS);
    for(int j = 0; j < OBSERVATIONS_SCATS; j++) {
      obs.getFeatures()(j, col) = 0; // initialization
    }
    int phase = (*i)->get_current_phase();
    obs.getFeatures()(phase, col) = 1; // 4 bits for current phase
    int next_entry = PHASES;
    
    double duration = (*i)->get_phase_duration(); // 1 double for current duration
    //cout << "sts: duration is " << duration << endl;
    obs.getFeatures()(next_entry, col) = duration;
    next_entry++;
    /*
    if(duration > 0)
      obs.getFeatures()(next_entry, col) = 1;
    next_entry++;
    if(duration > 1)
      obs.getFeatures()(next_entry, col) = 1;
    next_entry++;
    if(duration > 2)
      obs.getFeatures()(next_entry, col) = 1;
    next_entry++;
    */
    //for(int j = 0; j < PHASES; j++) {// 4 doubles for phase percentage history
    //obs.getFeatures()(next_entry, col) = (*i)->get_avg_phase_percent(j);
    //next_entry++;
    //}
    for(int d = 0; d < DIRECTIONS; d++) // 4 * 2 bits whether detector is active
      for(int q = 0; q < QUEUES; q++) {
	if((*i)->is_detector_active(d, q)) {
	  obs.getFeatures()(next_entry, col) = 1;
	  //cout << "detector["<< d <<"][" << q <<"] is active" << endl;
	}
	else {
	  //cout << "detector["<< d <<"][" << q <<"] is not active" << endl;
	}
	next_entry++;
      }
    
    for(int d = 0; d < DIRECTIONS; d++) {// 4 * 2  bits: past detector activity
      for(int q = 0; q < QUEUES; q++){
	obs.getFeatures()(next_entry, col) = 
	  (*i)->get_detector_count_last_cycle(d, q);
	next_entry++;
      }
    }
    for(int j = 0; j < PHASES; j++) { // 4 doubles: max possible throughput per phase
      obs.getFeatures()(next_entry, col) = 
	(*i)->get_max_throughput((Intersection::Phase) j); 
      next_entry++;
     }
    assert(next_entry == OBSERVATIONS_SCATS);
    //cout << "next entry " << next_entry << endl;
    /*
    if(Ps::pVERBOSE) { 
      for(int j = 0; j < OBSERVATIONS_SCATS; j++) 
	cout << " " << obs.getFeatures()(j, col); 
      cout << endl; 
      }
    */ 
    col++;
  }
  //cout << endl;
}
void STS::get_observation_for_mixed_controller(Observation& obs) {

  //int obs_learner = OBSERVATIONS_LEARNER;
  //if(Ps::pUSE_NEIGHBOUR_INFO) {
  //obs_learner = OBSERVATIONS_LEARNER_PLUS_NEIGHBOURS;
  //}
  int obs_learner = obsRows;
  //if(Ps::pCONTROLLER == _MixedController || 
  // Ps::pCONTROLLER == _MixedRPController ||
  // Ps::pCONTROLLER == _Scenario4Controller ) {
      obs_learner -= OBSERVATIONS_SCATS;
      //}

  int cols = intersections.size();
  //Observation scats_obs = new Observation();
  Observation scats_obs(OBSERVATIONS_SCATS, cols, cols);
  get_observation_for_scats_controller(scats_obs);
  project(obs.getFeatures(), range(0, OBSERVATIONS_SCATS), 
	  range(0, cols))  = scats_obs.getFeatures();

  Observation l_obs(obs_learner, cols, cols);
  get_observation_for_learning_controller(l_obs);
  project(obs.getFeatures(), range(OBSERVATIONS_SCATS, 
			      OBSERVATIONS_SCATS +
			      obs_learner), 
	  range(0, cols))  = l_obs.getFeatures();

  //column(scats_obs.getFeatures(), 0) = so;
}

void STS::update_reward(Car* car) {
  /*
  int travel_time = sim_time - car->get_create_time();
  //reward = reward + (1.0 / cars_arrived_last_time_step) * (-travel_time - reward);
  //reward_last_time_step -= travel_time;
  reward_last_time_step = -cars.size();
  if(Ps::pVERBOSE) {
    cout << "travel_time is " << travel_time;
    cout << ", cars_arrived_last_time_step is " << cars_arrived_last_time_step;
    cout << ", reward_last_time_step is " << reward_last_time_step << endl;
  }
  */
}

// Maximum episode length in steps. For planning mode. Only used
// by GOALPomdp at the moment.
int STS::getMaxEpisodeLength() {
    return 0;
}

// Get the number of rows in observation matrix. Generally the
// length of the observation vector.
int STS::getObsRows() {
    return obsRows;
}

// Get the number of columns in the observation matrix Typically
// use multiple columns to provide a different observation to each
// agent, so the number of columns is the same as getAgents()
int STS::getObsCols() {
    return obsCols;
}

// Get the total number of agents. Usually 1.
int STS::getAgents() {
    return agents;
}

// Get the total action dimensionality. Often, but not
// necessarily, the same as the number of agents.
int STS::getActionDim() {
    return actionDim;
}

int STS::getRewardDim() {
  if(Ps::pLOCAL_REWARDS) {
    return intersections.size();
  }
  else {
    return 1;
  }
}

int * STS::get_last_action()
{
	int j = 1;
	IntersectionVec::iterator inter;

	if (!last_action)
		last_action = new int [intersections.size() + 1];

// 	cout<<"In STS::get_last_action(), ";
	for (inter = intersections.begin(); inter < intersections.end() ; inter++) 
	{
		last_action[j ++] = (*inter)->get_current_phase() + 1;
//		cout<<last_action[j - 1]<<"  ";
	}
	
//	cout<<endl;
	
	return last_action;
}


char * STS::get_diagnosis()
{
	diagnosis_msg[0]= '\0';

	return diagnosis_msg;
}

