#ifndef CRFApprox_hh
#define CRFApprox_hh


namespace tree_mcmc {

class CRFApprox : public CRFBasic {

protected:

	int max_tree_iter;				// maximum iteration (for sampling stop)
	int max_tree_time;				// maximum time (for sampling stop)

protected:

	/** 
	 *	For CRF which is originally NOT a tree, then should apply MCMC sampler
	 */
	void build_init_tree(int root_id = -1);


	/**
	 *	Calculate the send, rec ... for all the trees in all the partitions
	 *	This is generic, once the trees and partitions are defined.
	 *	Let's skip the level of build_partition
	 */
	void build_all_trees();

	/**
	 *	Define how the graph is partitioned into trees.
	 */
	virtual void define_partitions() = 0;


	virtual void init_graph(int *argc, char **argv) = 0;

public:

	CRFApprox(int *argc, char** argv);

	// A generic sampler initializer
	virtual void init_tree_sampler(int *argc, char **argv);

	/**
	 *	Perform inference on the whole graph (tree) using RB
	 */
	virtual void infer_graph();


	/**
	 *	Perform sampling on the whole graph (tree) using RB
	 */
	virtual void sample_graph();



};

}

#endif
