#include "TreeBasics.hh"

using namespace std;

namespace tree_mcmc {

CRFChain::CRFChain(int *argc, char** argv):CRFExact(argc, argv)
{

}


void CRFChain::init_graph(int *argc, char **argv)
{
	Utility::get_arg(&global_num_node, argc, argv, "-na", 1, 0);
	num_row = 1;
	num_col = global_num_node;

	printf("In CRF for chain, num_agent = %d,\t num_class = %d\n", global_num_node, num_class);

	global_num_edge = global_num_node - 1;

	alloc_adj_matrix();
	alloc_potentials();
	associate_edge_with_nodes();
}


void CRFChain::alloc_adj_matrix()
{
	int i, j;

	adj_matrix = new int * [global_num_node + 1];
	for (i = 1; i <= global_num_node; i ++)
	{
		adj_matrix[i] = new int [global_num_node + 1];
		for (j = 1; j < i - 1; j ++)
			adj_matrix[i][j] = 0;
		if (i > 1)
			adj_matrix[i][i - 1] = 1;
		adj_matrix[i][i] = 0;
		if (i < global_num_node)
			adj_matrix[i][i + 1] = 1;
		for (j = i + 2; j <= global_num_node; j ++)
			adj_matrix[i][j] = 0;
	}

}


void CRFChain::associate_edge_with_nodes()
{
	int cur_item;
	Edge *p_edge;

	p_edge = edge_array + 1;
	for (cur_item = 1; cur_item < global_num_node; cur_item ++)
	{
		p_edge->p_node[1] = node_array + cur_item;
		p_edge->p_node[2] = node_array + cur_item + 1;
		p_edge ++;
	}
}


int CRFChain::get_edge_id(int node1, int node2)
{
	static int offset = global_num_node - num_row;

	if(node1 < node2)
	{
		if (node1 + 1 == node2)		// horizontal edge
			return node1 - int(node1 / (double)num_col);
		else						// vertical edge
			return offset + node1;
	}
	else
	{
		if (node2 + 1 == node1)		// horizontal edge
			return node2 - int(node2 / (double)num_col);
		else						// vertical edge
			return offset + node2;
	}	
}


void CRFChain::fill_edge_reward(double *reward)
{
	int cur_node;
	double *cur_edge_reward;

	cur_edge_reward = reward + global_num_node + 1;
	switch (edge_reward_def)
	{
	case 1:		// arithmetic mean of two end nodes (default)
		for (cur_node = 1; cur_node < num_col; cur_node ++)
			*(cur_edge_reward ++) = (reward[cur_node] + reward[cur_node + 1]) / 2.0;
		break;
	case 2:		// geometric mean
		for (cur_node = 1; cur_node < num_col; cur_node ++)
			*(cur_edge_reward ++) = sqrt(reward[cur_node] * reward[cur_node + 1]);
		break;
	case 3:		// min of two end nodes
		for (cur_node = 1; cur_node < num_col; cur_node ++)
			*(cur_edge_reward ++) = min(reward[cur_node], reward[cur_node + 1]);
		break;
	case 4:		// max of two end nodes
		for (cur_node = 1; cur_node < num_col; cur_node ++)
			*(cur_edge_reward ++) = max(reward[cur_node], reward[cur_node + 1]);
		break;
	case 5:		// end node [2]
		for (cur_node = 1; cur_node < num_col; cur_node ++)
			*(cur_edge_reward ++) = reward[cur_node + 1];
		break;
	case 6:		// end node [1]
		for (cur_node = 1; cur_node < num_col; cur_node ++)
			*(cur_edge_reward ++) = reward[cur_node];
		break;
	default:
		cout<<"Unforseen edge reward def: "<<edge_reward_def<<endl;
		exit(1);
	}

/*	
	for (cur_node = 1; cur_node <= num_col; cur_node ++)
		cout<<reward[cur_node]<<"  ";
	cout<<endl;
*/
}

}
