#include "TreeBasics.hh"


using namespace std;

namespace tree_mcmc {


CRFCross::CRFCross(int *argc, char** argv):CRFExact(argc, argv)
{

}

void CRFCross::init_graph(int *argc, char **argv)
{
	Utility::get_arg(&num_row, argc, argv, "-nr", 1, 0);
	Utility::get_arg(&num_col, argc, argv, "-nc", 1, 0);
	Utility::get_arg(&num_cross_upper, argc, argv, "-ncu", 1, 0);
	Utility::get_arg(&cross_hori_id, argc, argv, "-chi", 1, 0);

	Utility::get_arg(&global_num_node, argc, argv, "-na", 1, 0);

	printf("In CRF for cross, num_row = %d\tnum_col = %d\tnum_node = %d\tnum_class = %d\tnum_cross_upper = %d\tcross_hori_id = %d\n",
		num_row, num_col, global_num_node, num_class, num_cross_upper, cross_hori_id);
	
	if (global_num_node != num_col + num_row - 1)
	{
		cout<<"In init_graph_cross, global_num_node "<<global_num_node<<" is NOT equal to num_col + num_row - 1 ("
			<<num_col<<" + "<<num_row<<" - 1).  Abort."<<endl;
		exit(1);
	}

	global_num_edge = global_num_node - 1;	

	alloc_adj_matrix();
	alloc_potentials();
	associate_edge_with_nodes();
}


void CRFCross::alloc_adj_matrix()
{
 	int i, j, cur_item;
 	int *temp_id;
 
 	temp_id = new int [num_row + 1];
 	for (i = 1; i <= num_cross_upper; i ++)
 		temp_id[i] = i;
 	temp_id[i ++] = num_cross_upper + cross_hori_id;
 	cur_item = num_col + num_cross_upper + 1;
 	for (; i <= num_row; i ++)
 		temp_id[i] = cur_item ++;
 
 	adj_matrix = new int * [global_num_node + 1];
 	for (i = 1; i <= global_num_node; i ++)
 	{
 		adj_matrix[i] = new int [global_num_node + 1];
 		for (j = 1; j <= global_num_node; j ++)
 			adj_matrix[i][j] = 0;
 	}
 
 	adj_matrix[num_cross_upper + 1][num_cross_upper + 2] = 1;
 	for (i = num_cross_upper + 2; i < num_cross_upper + num_col; i ++)
 		adj_matrix[i][i - 1] = adj_matrix[i][i + 1] = 1;
 	adj_matrix[num_cross_upper + num_col][num_cross_upper + num_col - 1] = 1;
 
 	adj_matrix[temp_id[1]][temp_id[2]] = 1;
 	for (i = 2; i < num_row; i ++)
 		adj_matrix[temp_id[i]][temp_id[i - 1]] = adj_matrix[temp_id[i]][temp_id[i + 1]] = 1;
 	adj_matrix[temp_id[num_row]][temp_id[num_row - 1]] = 1;
 
 	delete []temp_id;	
}


void CRFCross::associate_edge_with_nodes()
{
	int cur_item, i;
	int *temp_id;
	Edge *p_edge;

	temp_id = new int [num_row + 1];
	for (i = 1; i <= num_cross_upper; i ++)
		temp_id[i] = i;
	temp_id[i ++] = num_cross_upper + cross_hori_id;
	cur_item = num_col + num_cross_upper + 1;
	for (; i <= num_row; i ++)
		temp_id[i] = cur_item ++;

	p_edge = edge_array + 1;
	for (cur_item = num_cross_upper + 1; cur_item < num_cross_upper + num_col; cur_item ++)
	{
		p_edge->p_node[1] = node_array + cur_item;
		p_edge->p_node[2] = node_array + cur_item + 1;
		p_edge ++;
	}

	p_edge = edge_array + num_col;
	for (cur_item = 1; cur_item < num_row; cur_item ++)
	{
		p_edge->p_node[1] = node_array + temp_id[cur_item];
		p_edge->p_node[2] = node_array + temp_id[cur_item + 1];
		p_edge ++;
	}

	delete []temp_id;
}


int CRFCross::get_edge_id(int node1, int node2)
{
	if (node1 > node2)
	{
		if (node2 <= num_cross_upper)
			return node2 + num_col - 1;	
		else if (node1 <= num_col + num_cross_upper)
			return node2 - num_cross_upper;
		return node1 - 1;
	}
	else
	{
		if (node1 <= num_cross_upper)
			return node1 + num_col - 1;	
		else if (node2 <= num_col + num_cross_upper)
			return node1 - num_cross_upper;
		return node2 - 1;
	}
}


void CRFCross::fill_edge_reward(double *reward)
{
	int cur_edge;
	double *cur_reward;
	Edge *pEdge;

	cur_reward = reward + global_num_node + 1;
	pEdge = edge_array + 1;

	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		switch (edge_reward_def)
		{
		case 1:		// arithmetic mean of two end nodes (default)
			*(cur_reward ++) = (reward[pEdge->p_node[1] - node_array] + reward[pEdge->p_node[2] - node_array]) / 2;
			break;
		case 2:		// geometric mean
			*(cur_reward ++) = sqrt(reward[pEdge->p_node[1] - node_array] * reward[pEdge->p_node[2] - node_array]);
			break;
		case 3:		// min of two end nodes
			*(cur_reward ++) = min(reward[pEdge->p_node[1] - node_array], reward[pEdge->p_node[2] - node_array]);
			break;
		case 4:		// max of two end nodes
			*(cur_reward ++) = max(reward[pEdge->p_node[1] - node_array], reward[pEdge->p_node[2] - node_array]);
			break;
		case 5:		// end node [2]
			*(cur_reward ++) = reward[pEdge->p_node[2] - node_array];
			break;
		case 6:		// end node [1]
			*(cur_reward ++) = reward[pEdge->p_node[1] - node_array];
			break;
		default:
			cout<<"Unforseen edge reward def: "<<edge_reward_def<<endl;
			exit(1);
		}

		pEdge ++;
	}
}

}
