#include "TreeBasics.hh"

using namespace std;

namespace tree_mcmc {

CRFGrid::CRFGrid(int *argc, char** argv):CRFApprox(argc, argv)
{

}


void CRFGrid::init_graph(int *argc, char **argv)
{
	Utility::get_arg(&num_row, argc, argv, "-nr", 1, 0);
	Utility::get_arg(&num_col, argc, argv, "-nc", 1, 0);
	Utility::get_arg(&global_num_node, argc, argv, "-na", 1, 0);
	
	printf("In CRF for grid, num_row = %d\tnum_col = %d\tnum_class = %d\n", num_row, num_col, num_class);

	if (global_num_node != num_col * num_row)
	{
		cout<<"In init_graph_grid, global_num_node "<<global_num_node<<" is NOT equal to num_col * num_row ("
			<<num_col<<" * "<<num_row<<").  Abort."<<endl;
		exit(1);
	}
	
	global_num_edge = 2 * global_num_node - num_row - num_col;	

	alloc_adj_matrix();
	alloc_potentials();
	associate_edge_with_nodes();
	define_partitions();
}


void CRFGrid::alloc_adj_matrix()
{
	int cur_row, cur_col, i, j, cur_node, *temp;

	adj_matrix = new int * [global_num_node + 1];
	for (i = 1; i <= global_num_node; i ++)
	{
		adj_matrix[i] = new int [global_num_node + 1];
		temp = adj_matrix[i] + 1;
		for (j = 1; j <= global_num_node; j ++)
			*(temp++) = 0;
	}

	cur_node = 1;
	for (cur_row = 1; cur_row <= num_row; cur_row ++)
	{
		for (cur_col = 1; cur_col < num_col; cur_col ++)
		{
			adj_matrix[cur_node][cur_node + 1] = 1;
			adj_matrix[cur_node + 1][cur_node] = 1;
			cur_node ++;
		}
		cur_node ++;   
	}

	cur_node = 1;
	i = num_col + 1;
	for (cur_row = 1; cur_row < num_row; cur_row ++)
	{
		for (cur_col = 1; cur_col <= num_col; cur_col ++)
		{
			adj_matrix[cur_node][i] = 1;
			adj_matrix[i++][cur_node++] = 1;
		}
	}

/*
	for (cur_row = 1; cur_row <= global_num_node; cur_row ++)
	{
		for (cur_col = 1; cur_col <= global_num_node; cur_col ++)
		{
			printf("%d\t", adj_matrix[cur_row][cur_col]);	
		}
		printf("\n");
	}
*/
}


void CRFGrid::associate_edge_with_nodes()
{
	int cur_node, cur_row, cur_col, cur_node_2;
	Edge *p_edge;

	p_edge = edge_array + 1;

	// Horizontal edges
	cur_node = 1;
	for (cur_row = 1; cur_row <= num_row; cur_row ++)
	{
		for (cur_col = 1; cur_col < num_col; cur_col ++)
		{
			p_edge->p_node[1] = node_array + cur_node;
			cur_node ++;
			p_edge->p_node[2] = node_array + cur_node;			
			p_edge ++;
		}
		cur_node ++;
	}

	// Vertical edges
	cur_node = 1;
	cur_node_2 = num_col + 1;
	for (cur_row = 1; cur_row < num_row; cur_row ++)
	{
		for (cur_col = 1; cur_col <= num_col; cur_col ++)
		{
			p_edge->p_node[1] = node_array + cur_node++;
			p_edge->p_node[2] = node_array + cur_node_2++;			
			p_edge ++;
		}
	}
}


int CRFGrid::get_edge_id(int node1, int node2)
{
	static int offset = global_num_node - num_row;

	if(node1 < node2)
	{
		if (node1 + 1 == node2)		// horizontal edge
			return node1 - int(node1 / (double)num_col);
		else						// vertical edge
			return offset + node1;
	}
	else
	{
		if (node2 + 1 == node1)		// horizontal edge
			return node2 - int(node2 / (double)num_col);
		else						// vertical edge
			return offset + node2;
	}	
}


/**
 *	Construct trees in partitions
 *	ONLY determines the id of the nodes in the trees
 */
void CRFGrid::define_partitions()
{
	int cur_part, cur_row, cur_col, next_index, cur_id, start_id, single;
	Tree *p_tree;
	Partition *p_part;

	global_num_partition = 4;
	partitions = new Partition[5];

	for (cur_part = 1;  cur_part <= 4; cur_part ++)		// for the 4 partitions
	{
		p_part = partitions + cur_part;
		p_part->num_trees = 2;
		p_part->trees = new Tree[3];

#ifdef RB_NODE_SINGLE
		if (cur_part == 1)
		{
			p_part->trees[1].cal_node_rb = 1;
			p_part->trees[2].cal_node_rb = 1;
		}
		else
		{
			p_part->trees[1].cal_node_rb = 0;
			p_part->trees[2].cal_node_rb = 0;
		}
#else
		p_part->trees[1].cal_node_rb = 1;
		p_part->trees[2].cal_node_rb = 1;
#endif
	
		if (cur_part == 1)
		{
			// first upper tree
			p_tree = p_part->trees + 1;
			p_tree->num_node = int((num_col + 0.1) / 2.0) * (num_row - 2) + num_col;
			p_tree->local_to_global = new int[p_tree->num_node + 1];
						
			for (next_index = 1; next_index <= num_col; next_index++)
				p_tree->local_to_global[next_index] = next_index;

			start_id = cur_id = num_col + 2;
			for (cur_row = 2; cur_row < num_row; cur_row ++)
			{
				for (cur_col = 2; cur_col <= num_col; cur_col += 2)
				{
					p_tree->local_to_global[next_index++] = cur_id;
					cur_id += 2;
				}
				start_id += num_col;
				cur_id = start_id;
			}

			// second lower tree
			p_tree = p_part->trees + 2;
			p_tree->num_node = int((num_col + 1.1) / 2) * (num_row - 2) + num_col;
			p_tree->local_to_global = new int[p_tree->num_node + 1];

			next_index = 1;
			start_id = cur_id = num_col + 1;
			for (cur_row = 2; cur_row < num_row; cur_row ++)
			{
				for (cur_col = 1; cur_col <= num_col; cur_col += 2)
				{
					p_tree->local_to_global[next_index++] = cur_id;
					cur_id += 2;
				}
				start_id += num_col;
				cur_id = start_id;
			}

			for (cur_col = 1; cur_col <= num_col; cur_col ++)
				p_tree->local_to_global[next_index++] = cur_id ++;
		}

		else if (cur_part == 2)
		{
			// first upper tree
			p_tree = p_part->trees + 1;
			p_tree->num_node = int((num_col + 1.1) / 2) * (num_row - 2) + num_col;
			p_tree->local_to_global = new int[p_tree->num_node + 1];
						
			for (next_index = 1; next_index <= num_col; next_index++)
				p_tree->local_to_global[next_index] = next_index;

			start_id = cur_id = num_col + 1;
			for (cur_row = 2; cur_row < num_row; cur_row ++)
			{
				for (cur_col = 1; cur_col <= num_col; cur_col += 2)
				{
					p_tree->local_to_global[next_index++] = cur_id;
					cur_id += 2;
				}
				start_id += num_col;
				cur_id = start_id;
			}

			// second lower tree
			p_tree = p_part->trees + 2;
			p_tree->num_node = int((num_col + 0.1) / 2.0) * (num_row - 2) + num_col;
			p_tree->local_to_global = new int[p_tree->num_node + 1];

			next_index = 1;
			start_id = cur_id = num_col + 2;
			for (cur_row = 2; cur_row < num_row; cur_row ++)
			{
				for (cur_col = 2; cur_col <= num_col; cur_col += 2)
				{
					p_tree->local_to_global[next_index++] = cur_id;
					cur_id += 2;
				}
				start_id += num_col;
				cur_id = start_id;
			}

			cur_id --;
			for (cur_col = 1; cur_col <= num_col; cur_col ++)
				p_tree->local_to_global[next_index++] = cur_id ++;
		}

		else if (cur_part == 3)
		{
			// first left tree
			p_tree = p_part->trees + 1;
			p_tree->num_node = int((num_row + 0.1) / 2.0) * (num_col - 2) + num_row;
			p_tree->local_to_global = new int[p_tree->num_node + 1];

			single = 1;
			next_index = 1;
			cur_id = 1;

			for (cur_row = 1; cur_row <= num_row; cur_row ++)
			{
				if (single)
				{
					p_tree->local_to_global[next_index++] = cur_id;
					cur_id += num_col;
				}
				else
				{
					for (cur_col = 1; cur_col < num_col; cur_col ++)
						p_tree->local_to_global[next_index++] = cur_id++;
					cur_id ++;
				}

				single = !single;
			}

			// second right tree
			p_tree = p_part->trees + 2;
			p_tree->num_node = int((num_row + 1.1) / 2) * (num_col - 2) + num_row;
			p_tree->local_to_global = new int[p_tree->num_node + 1];

			single = 0;
			next_index = 1;
			cur_id = 2;

			for (cur_row = 1; cur_row <= num_row; cur_row ++)
			{
				if (single)
				{
					p_tree->local_to_global[next_index++] = cur_id;
					cur_id += 2;
				}
				else
				{					
					for (cur_col = 2; cur_col <= num_col; cur_col ++)
						p_tree->local_to_global[next_index++] = cur_id++;
					cur_id += num_col - 1;
				}
				
				single = !single;
			}			
		}

		else if (cur_part == 4)
		{
			// first left tree
			p_tree = p_part->trees + 1;
			p_tree->num_node = int((num_row + 1.1) / 2) * (num_col - 2) + num_row;
			p_tree->local_to_global = new int[p_tree->num_node + 1];

			single = 0;
			next_index = 1;
			cur_id = 1;

			for (cur_row = 1; cur_row <= num_row; cur_row ++)
			{
				if (single)
				{
					p_tree->local_to_global[next_index++] = cur_id;
					cur_id += num_col;
				}
				else
				{
					for (cur_col = 1; cur_col < num_col; cur_col ++)
						p_tree->local_to_global[next_index++] = cur_id++;
					cur_id ++;
				}

				single = !single;
			}

			// second right tree
			p_tree = p_part->trees + 2;
			p_tree->num_node = int((num_row + 0.1) / 2.0) * (num_col - 2) + num_row;
			p_tree->local_to_global = new int[p_tree->num_node + 1];

			single = 1;
			next_index = 1;
			cur_id = num_col;

			for (cur_row = 1; cur_row <= num_row; cur_row ++)
			{
				if (single)
				{
					p_tree->local_to_global[next_index++] = cur_id;
					cur_id += 2;
				}
				else
				{					
					for (cur_col = 2; cur_col <= num_col; cur_col ++)
						p_tree->local_to_global[next_index++] = cur_id++;
					cur_id += num_col - 1;
				}

				single = !single;
			}			
		}
	}
}


void CRFGrid::fill_edge_reward(double *reward)
{
	int cur_row, cur_col, cur_node, cur_node_2;
	double *cur_edge_reward;

	// Horizontal edges
	cur_edge_reward = reward + global_num_node + 1;
	cur_node = 1;
	for (cur_row = 1; cur_row <= num_row; cur_row ++)
	{
		switch (edge_reward_def)
		{
		case 1:		// arithmetic mean of two end nodes (default)
			for (cur_col = 1; cur_col < num_col; cur_col ++)
			{
				*(cur_edge_reward ++) = (reward[cur_node] + reward[cur_node + 1]) / 2.0;			
				cur_node ++;
			}			
			break;
		case 2:		// geometric mean
			for (cur_col = 1; cur_col < num_col; cur_col ++)
			{
				*(cur_edge_reward ++) = sqrt(reward[cur_node] * reward[cur_node + 1]);
				cur_node ++;
			}			
			break;
		case 3:		// min of two end nodes
			for (cur_col = 1; cur_col < num_col; cur_col ++)
			{
				*(cur_edge_reward ++) = min(reward[cur_node], reward[cur_node + 1]);
				cur_node ++;
			}			
			break;
		case 4:		// max of two end nodes
			for (cur_col = 1; cur_col < num_col; cur_col ++)
			{
				*(cur_edge_reward ++) = max(reward[cur_node], reward[cur_node + 1]);
				cur_node ++;
			}			
			break;
		case 5:		// end node [2]
			for (cur_col = 1; cur_col < num_col; cur_col ++)
			{
				*(cur_edge_reward ++) = reward[cur_node + 1];
				cur_node ++;
			}			
			break;
		case 6:		// end node [1]
			for (cur_col = 1; cur_col < num_col; cur_col ++)
			{
				*(cur_edge_reward ++) = reward[cur_node];
				cur_node ++;
			}
			break;
		default:
			cout<<"Unforseen edge reward def: "<<edge_reward_def<<endl;
			exit(1);
		}

		cur_node ++;
	}

	// Vertical edges
	cur_node = 1;
	cur_node_2 = num_col + 1;
	for (cur_row = 1; cur_row < num_row; cur_row ++)
	{
		switch (edge_reward_def)
		{
		case 1:		// arithmetic mean of two end nodes (default)
			for (cur_col = 1; cur_col <= num_col; cur_col ++)
				*(cur_edge_reward ++) = (reward[cur_node ++] + reward[cur_node_2 ++]) / 2.0;
			break;
		case 2:		// geometric mean
			for (cur_col = 1; cur_col <= num_col; cur_col ++)
				*(cur_edge_reward ++) = sqrt(reward[cur_node ++] * reward[cur_node_2 ++]);
			break;
		case 3:		// min of two end nodes
			for (cur_col = 1; cur_col <= num_col; cur_col ++)
				*(cur_edge_reward ++) = min(reward[cur_node ++], reward[cur_node_2 ++]);
			break;
		case 4:		// max of two end nodes
			for (cur_col = 1; cur_col <= num_col; cur_col ++)
			{
				*(cur_edge_reward ++) = max(reward[cur_node ++], reward[cur_node_2 ++]);
				cur_node ++;			cur_node_2 ++;
			}
			break;
		case 5:		// end node [2]
			for (cur_col = 1; cur_col <= num_col; cur_col ++)
				*(cur_edge_reward ++) = reward[cur_node_2 ++];
			break;
		case 6:		// end node [1]
			for (cur_col = 1; cur_col <= num_col; cur_col ++)
				*(cur_edge_reward ++) = reward[cur_node ++];
			break;
		default:
			cout<<"Unforseen edge reward def: "<<edge_reward_def<<endl;
			exit(1);
		}
	}
}





}
