#include "TreeBasics.hh"


using namespace std;

namespace tree_mcmc {


/** 
 *	For CRF which is a chain, then work on this tree
 */

CRFOffset::CRFOffset(int *argc, char** argv):CRFChain(argc, argv)
{
	sum_act_3 = 0;
	
	Utility::get_arg(&max_delay, argc, argv, "-xdy", 1, 0);
	Utility::get_arg(&min_delay, argc, argv, "-idy", 1, 0);
	Utility::get_arg(&edge_fea_sel, argc, argv, "-efs", 1, 0);

	cout<<"In CRFOffset, max_delay = "<<max_delay<<", min_delay = "<<min_delay<<", edge_fea_sel = "<<edge_fea_sel<<endl;
}


/**
 *	After calculating the RB estimations on the graph,
 *	Calculate the feature expectations.
 */
void CRFOffset::cal_fea_expectation()
{
	int cur_edge, cur_class_1, cur_class_2, cur_delay, cur_index, rec_start, sample;
	int node1, node2;
	double *rb1, *rb1_2, *fept1, **fept2, **cob2;
	Edge *p_edge;

	cal_node_fea_exp();

	// Second, calculate the edge feature expectation
	p_edge = edge_array + 1;
	fept2 = edge_fea_exp + 1;

	rec_start = cur_step % max_delay;
	if(!rec_start)
		rec_start = max_delay;
	
// 	cout<<"In CRFOffset::cal_fea_expectation(), ";
// 	for (int j = 1; j <= global_num_node; j ++)
// 		cout<<history_action[j][rec_start]<<"  ";
// 	cout<<endl;

	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)		// loop over edges
	{
		fept1 = *(fept2 ++) + 1;

		if (edge_fea_sel & 1)		// Using binary constant edge bias
		{
			*fept1 = 0.0;

			for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
				 *fept1 += p_edge->rb_est[cur_class_1][cur_class_1];
			fept1 ++;
		}

		if (edge_fea_sel & 2)		// Using square matrix constant edge bias
		{
			cob2 = p_edge->rb_est;
			for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
				for (cur_class_2 = 1; cur_class_2 <= num_class; cur_class_2 ++)
					*(fept1 ++) = cob2[cur_class_1][cur_class_2];
		}

		if (cur_step > max_delay && (edge_fea_sel & 60))
		{
			rb1 = p_edge->p_node[1]->rb_est;
			rb1_2 = p_edge->p_node[2]->rb_est;
			node1 = p_edge->p_node[1] - node_array;
			node2 = p_edge->p_node[2] - node_array;

			cur_index = rec_start;
			for (cur_delay = min_delay; cur_delay <= max_delay; cur_delay ++, cur_index ++)
			{
				if (edge_fea_sel & 4)		// First, 1 past, 2 current (ONE feature)					
					*(fept1 ++) = rb1_2[history_action[node1][cur_index]];
				
				if (edge_fea_sel & 8)		// Second, 1 past, 2 current (square matrix)
				{						
					// First, 1 past, 2 current
					sample = history_action[node1][cur_index];	
					fept1 += (sample - 1) * num_class;
					for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
						*(fept1 ++) = rb1_2[cur_class_1];
					fept1 += (num_class - sample) * num_class;
				}

				if (edge_fea_sel & 16)		// Third, 1 current, 2 past (ONE feature)
					*(fept1 ++) = rb1[history_action[node2][cur_index]];

				if (edge_fea_sel & 32)		// Fourth, 1 current, 2 past (square matrix)
				{
					// Second, 1 current, 2 past
					sample = history_action[node2][cur_index];
					fept1 += (sample - 1) * num_class;
					for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
						*(fept1 ++) = rb1[cur_class_1];
					fept1 += (num_class - sample) * num_class;
				}

				if (cur_index == max_delay)
					cur_index = 0;
			}
		}
		p_edge ++;
	}
}


void CRFOffset::fill_trace()
{
	int cur_item, cur_fea, cur_class, node1, node2, cur_index, rec_start, cur_delay, sample1, sample2, offset;
	double **pd2, *pd1, **pd2_2, *pd1_2;
	Edge *p_edge;

	fill_node_trace();

	pd2 = edge_fea_trace + 1;
	pd2_2 = edge_fea_exp + 1;
	p_edge = edge_array + 1;

	rec_start = cur_step % max_delay;
	if(!rec_start)
		rec_start = max_delay;
	offset = num_class * num_class + 1;

	for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
	{
		pd1 = *(pd2 ++) + 1;
		pd1_2 = *(pd2_2 ++) + 1;

		if (edge_fea_sel & 1)		// Using binary constant edge bias
		{
			if (p_edge->p_node[1]->sampled_value == p_edge->p_node[2]->sampled_value)
			{
				*pd1_2 = 1.0 - *pd1_2;
				*(pd1 ++) += *(pd1_2 ++);
			}
			else
			{
				*pd1_2 = - *pd1_2;
				*(pd1++) += *(pd1_2 ++);
			}
		}

		if (edge_fea_sel & 2)		// Using square matrix constant edge bias
		{
			for (cur_fea = 1; cur_fea < offset; cur_fea ++)			
			{
				*pd1_2 = - *pd1_2;
				*(pd1 ++) += *(pd1_2 ++);
			}

			cur_class = (p_edge->p_node[1]->sampled_value - 1) * num_class
							+ p_edge->p_node[2]->sampled_value - offset;

			*(pd1 + cur_class) += 1;
			*(pd1_2 + cur_class) += 1;
		}

		if (cur_step > max_delay && (edge_fea_sel & 60))
		{
			node1 = p_edge->p_node[1] - node_array;
			node2 = p_edge->p_node[2] - node_array;				

			cur_index = rec_start;
			for (cur_delay = min_delay; cur_delay <= max_delay; cur_delay ++, cur_index ++)
			{
				if (edge_fea_sel & 4)		// First, 1 past, 2 current (ONE feature)	
				{
					if (p_edge->p_node[2]->sampled_value == history_action[node1][cur_index])
					{
						*pd1_2 = 1.0 - *pd1_2;				
						*(pd1 ++) += *(pd1_2 ++);
					}
					else
					{
						*pd1_2 = - *pd1_2;
						*(pd1++) += *(pd1_2 ++);				
					}
				}

				if (edge_fea_sel & 8)		// Second, 1 past, 2 current (square matrix)
				{
					sample1 = history_action[node1][cur_index];
					pd1 += (sample1 - 1) * num_class;
					pd1_2 += (sample1 - 1) * num_class;
					for (cur_class = 1; cur_class <= num_class; cur_class ++)
					{
						*pd1_2 = - *pd1_2;
						*(pd1 ++) += *(pd1_2 ++);
					}

					sample2 = p_edge->p_node[2]->sampled_value - num_class - 1;
					*(pd1_2 + sample2) += 1;
					*(pd1 + sample2) += 1;

					pd1 += (num_class - sample1) * num_class;
					pd1_2 += (num_class - sample1) * num_class;
				}

				if (edge_fea_sel & 16)		// Third, 1 current, 2 past (ONE feature)
				{
					if (p_edge->p_node[1]->sampled_value == history_action[node2][cur_index])
					{
						*pd1_2 = 1.0 - *pd1_2;				
						*(pd1 ++) += *(pd1_2 ++);
					}
					else
					{
						*pd1_2 = - *pd1_2;
						*(pd1++) += *(pd1_2 ++);				
					}
				}

				if (edge_fea_sel & 32)		// Fourth, 1 current, 2 past (square matrix)
				{
					sample1 = history_action[node2][cur_index];
					pd1 += (sample1 - 1) * num_class;
					pd1_2 += (sample1 - 1) * num_class;
					for (cur_class = 1; cur_class <= num_class; cur_class ++)
					{
						*pd1_2 = - *pd1_2;
						*(pd1 ++) += *(pd1_2 ++);
					}

					sample2 = p_edge->p_node[1]->sampled_value - num_class - 1;
					*(pd1_2 + sample2) += 1;
					*(pd1 + sample2) += 1;

					pd1 += (num_class - sample1) * num_class;
					pd1_2 += (num_class - sample1) * num_class;
				}

				if (cur_index == max_delay)
					cur_index = 0;
			}
		}
		p_edge ++;
	}
}


/**
 *	Set the edge and node potentials of the graph.  
 *	Space allocation has been done by alloc_potentials()
 */
void CRFOffset::set_potentials()
{
	int cur_item, cur_class_1, cur_class_2, rec_start, cur_index, cur_delay, node1, node2, sample, offset;
	Edge *p_edge;
	double *cur_pot1, *cur_pot2, **efp2, *efp1;

	set_node_potentials();

	//////////////////////////////////////////////////////////////////////////
	//  Edge potentials

	p_edge = edge_array + 1;
	efp2 = edge_fea_para + 1;

	rec_start = cur_step % max_delay;
	if(!rec_start)
		rec_start = max_delay;

// 	cout<<"In CRFOffset::set_potentials(), ";
// 
// 	for (int j=1; j <= global_num_node; j ++)
// 	{
// 		cout<<history_action[j][rec_start]<<"  ";
// 	}
// 	cout<<endl;

	for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)		// loop over edges
	{
		efp1 = *(efp2 ++) + 1;

		if (edge_fea_sel & 1)
		{
			for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
				for (cur_class_2 = 1; cur_class_2 <= num_class; cur_class_2 ++)
				{
					p_edge->edge_pot[1][cur_class_1][cur_class_2] = *(efp1 ++);
				}
			for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
				for (cur_class_2 = 1; cur_class_2 <= num_class; cur_class_2 ++)
				{
					p_edge->edge_pot[2][cur_class_1][cur_class_2] = p_edge->edge_pot[1][cur_class_2][cur_class_1];
				}
		}

		if (edge_fea_sel & 2)
		{
			if (edge_fea_sel & 1)
			{
				for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
				{
					p_edge->edge_pot[1][cur_class_1][cur_class_1] += (*efp1);
					p_edge->edge_pot[2][cur_class_1][cur_class_1] = p_edge->edge_pot[1][cur_class_1][cur_class_1];
				}
			}
			else		// change += to =
			{
				for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
				{
					p_edge->edge_pot[1][cur_class_1][cur_class_1] = (*efp1);
					p_edge->edge_pot[2][cur_class_1][cur_class_1] = p_edge->edge_pot[1][cur_class_1][cur_class_1];
				}
			}
			efp1 ++;
		}

		if (cur_step > max_delay && (edge_fea_sel & 60))
		{
			node1 = p_edge->p_node[1] - node_array;
			node2 = p_edge->p_node[2] - node_array;
			cur_pot1 = p_edge->p_node[1]->node_pot + 1;
			cur_pot2 = p_edge->p_node[2]->node_pot + 1;

			cur_index = rec_start;
			for (cur_delay = min_delay; cur_delay <= max_delay; cur_delay ++, cur_index ++)
			{
				if (edge_fea_sel & 4)		// First, 1 past, 2 current (ONE feature)					
					cur_pot2[history_action[node1][cur_index]] += *(efp1 ++);

				if (edge_fea_sel & 8)		// Second, 1 past, 2 current (square matrix)
				{			
					sample = history_action[node1][cur_index];
					efp1 += (sample - 1) * num_class;
					for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
						*(cur_pot2 ++) += *(efp1 ++);
					cur_pot2 -= num_class;
					efp1 += (num_class - sample) * num_class;
				}

				if (edge_fea_sel & 16)		// Third, 1 current, 2 past (ONE feature)
					cur_pot1[history_action[node2][cur_index]] += *(efp1 ++);

				if (edge_fea_sel & 32)		// Fourth, 1 current, 2 past (square matrix)
				{				
					sample = history_action[node2][cur_index];
					efp1 += (sample - 1) * num_class;
					for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
						*(cur_pot1 ++) += *(efp1 ++);
					cur_pot1 -= num_class;
					efp1 += (num_class - sample) * num_class;
				}

				if (cur_index == max_delay)
					cur_index = 0;
			}
		}

		p_edge ++;
	}
}


void CRFOffset::init_tree_sampler(int *argc, char **argv)
{
	CRFChain::init_tree_sampler(argc, argv);

	Utility::get_arg(&cycleTime, argc, argv, "-cyt", 0, -1);
	Utility::get_arg(&bias, argc, argv, "-bias", 1, 0);

	if (fabs(bias) > 1e-5)
	{
		if (cycleTime % num_class != 0)
		{
			cout<<"!! CyclicPolicyBias requires cycleTime="
				<<cycleTime
				<<" have outputs="
				<<num_class
				<<" as a factor\n";
			exit(EXIT_FAILURE);
		}
		
		stepsPerControl = cycleTime / num_class;
		cout<<"Bias = "<<bias<<"\tstepsPerControl = "<<stepsPerControl<<"\tcycleTime = "<<cycleTime<<endl;
	}

	if (edge_fea_sel & 60)
	{
		history_action = new int * [global_num_node + 1];
		for (int cur_node = 1; cur_node <= global_num_node; cur_node ++)
			history_action[cur_node] = new int [max_delay + 1];
	}
}


void CRFOffset::cal_diagnosis()
{
	int cur_node;
	for(cur_node = 2; cur_node < global_num_node; cur_node ++)
	{
		if (node_array[cur_node].sampled_value != 3)
			sum_act_3 += 1;	
	}
}


void CRFOffset::load_last_action()
{
	if (edge_fea_sel & 60)
	{
		int cur_item, rec_start;

		rec_start = cur_step % max_delay;
		if(!rec_start)
			rec_start = max_delay;
				
		for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
			history_action[cur_item][rec_start] = last_action[cur_item];
	}
}


void CRFOffset::buffer_last_action(int * actions)
{
	if (edge_fea_sel & 60)
	{
		for (int j = 1; j <= global_num_node; j ++) 
			last_action[j] = actions[j];
	}
}


char * CRFOffset::get_diagnosis(int argc, char ** argv)
{	
	int step_per_epoch;

	Utility::get_arg(&step_per_epoch, &argc, argv, "-spe", 1, 0);
	sprintf(message, "%s : %g", string_max_param, sum_act_3 / (double) step_per_epoch);

	sum_act_3 = 0;

	return message;
}


void CRFOffset::graph_para_write(std::ostream & o) 
{
	int cur_node, cur_fea, cur_class, cur_edge, cur_obs_dim;
	double ***nfp3, **nfp2, *nfp1, **efp2, *efp1;
	Node *p_node;

	//  First, nodes
	nfp3 = node_fea_para + 1;
	p_node = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		nfp2 = *(nfp3 ++) + 1;
		for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
			{
				// node_fea_para[cur_node][cur_class][cur_obs_dim];
				o << *(nfp1 ++) << "  ";	 // update node feature grads, node_fea_para[i][j][k]
			}
			o << endl;
		}
		o << endl;

		p_node ++;
	}

	// If possible, extra node features
	if (num_node_extra_fea)
	{
		nfp2 = node_extra_fea_para + 1;		
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_class = 1; cur_class <= num_node_extra_fea; cur_class ++)
				o << *(nfp1 ++) << "  ";
			o << endl;
		}
	}

	//  Second, edges
	efp2 = edge_fea_para + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		efp1 = *(efp2 ++) + 1;
		for (cur_class = 1; cur_class <= num_class; cur_class ++)
		{
			for (cur_fea = 1; cur_fea <= num_class; cur_fea ++)
				o << *(efp1 ++) <<"  ";
			o<<endl;
		}

		o << endl;
	}
}


}
