#include "TreeBasics.hh"


using namespace std;

namespace tree_mcmc {


/** 
 *	For CRF which is a chain, then work on this tree
 */

CRFSensorCircle::CRFSensorCircle(int *argc, char** argv):CRFCircle(argc, argv)
{

}

/**
 *	After calculating the RB estimations on the graph,
 *	Calculate the feature expectations.
 */
void CRFSensorCircle::cal_fea_expectation()
{
	int cur_edge, cur_node;	
	double *rb1, *fept1, **fept2, temp;
	Node *p_node;
	Edge *p_edge;

	// Node features
	cal_node_fea_exp();

	// Node Extra features
	fept2 = node_extra_fea_exp + 1;
	p_node = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		rb1 = p_node->rb_est + 1;
		*(*(fept2 ++) + 1) = *(rb1 + 1) + *(rb1 + 2);
		p_node ++;
	}

	// Edge features
	// There are 3 edge (a, b) features: 
	// 1) both are focusing on the cell in the middle a right, b left
	// 2) one and only one is focusing on the cell
	// 3) none is focusing on the cell

	p_edge = edge_array + 1;
	fept2 = edge_fea_exp + 1;

	// 1st, upper and lower row
	for (cur_edge = 2; cur_edge < global_num_edge; cur_edge ++)
	{
		fept1 = * (fept2 ++) + 3;

		temp = p_edge->rb_est[3][2] * edge_fea_intensity;
		*(fept1 --) = temp;
		*fept1 = (p_edge->rb_est[1][2] + p_edge->rb_est[2][2] 
						+ p_edge->rb_est[3][1] + p_edge->rb_est[3][3]) * edge_fea_intensity;

		temp += *(fept1 --);
		*fept1 = edge_fea_intensity - temp;

		p_edge ++;
	}

	// 2nd, the left end top-bottom pair
	fept1 = * (fept2 ++) + 3;

	temp = p_edge->rb_est[3][3] * edge_fea_intensity;
	*(fept1 --) = temp;
	*fept1 = (p_edge->rb_est[1][3] + p_edge->rb_est[2][3] 
					+ p_edge->rb_est[3][1] + p_edge->rb_est[3][2]) * edge_fea_intensity;
	temp += *(fept1 --);
	*fept1 = edge_fea_intensity - temp;
	p_edge ++;

	// 3rd, the right end top-bottom pair
	fept1 = * (fept2 ++) + 3;

	temp = p_edge->rb_est[2][2] * edge_fea_intensity;
	*(fept1 --) = temp;
	*fept1 = (p_edge->rb_est[1][2] + p_edge->rb_est[2][1] 
					+ p_edge->rb_est[2][3] + p_edge->rb_est[3][2]) * edge_fea_intensity;
	temp += *(fept1 --);
	*fept1 = edge_fea_intensity - temp;
}


void CRFSensorCircle::fill_trace()
{
	int cur_item, sample1, sample2;
	double **pd2, *pd1, **pd2_2, *pd1_2;
	bool flag;
	Node *p_node;
	Edge *p_edge;

	// Node features
	fill_node_trace();

	// Node Extra features
	p_node = node_array + 1;
	pd2 = node_extra_fea_trace + 1;
	pd2_2 = node_extra_fea_exp + 1;
	for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
	{
		pd1_2 = *(pd2_2 ++) + 1;

		*pd1_2 = - *pd1_2;
		if (p_node->sampled_value > 1)
			*pd1_2 += 1;

		*(*(pd2 ++) + 1) += *pd1_2;

		p_node ++;
	}
	
	// Edge features
	pd2 = edge_fea_trace + 1;
	pd2_2 = edge_fea_exp + 1;
	p_edge = edge_array + 1;

	// 1st: upper and lower row
	for (cur_item = 2; cur_item < global_num_node; cur_item ++)
	{
		pd1 = *(pd2 ++) + 3;
		pd1_2 = *(pd2_2 ++) + 3;
		flag = true;

		sample1 = p_edge->p_node[1]->sampled_value;
		sample2 = p_edge->p_node[2]->sampled_value;

		// Feature 2
		if (sample1 == 3 && sample2 == 2)
		{
			flag = false;
			*pd1_2 = edge_fea_intensity - *pd1_2;				
			*(pd1--) += *(pd1_2 --);
		}
		else
		{
			*pd1_2 = - *pd1_2;
			*(pd1--) += *(pd1_2 --);
		}

		// Feature 1
		if (flag && (sample1 == 1 && sample2 == 2 || sample1 == 2 && sample2 == 2 ||
					sample1 == 3 && sample2 == 1 || sample1 == 3 && sample2 == 3))
		{
			flag = false;
			*pd1_2 = edge_fea_intensity - *pd1_2;				
			*(pd1--) += *(pd1_2 --);
		}
		else
		{
			*pd1_2 = - *pd1_2;
			*(pd1--) += *(pd1_2 --);
		}

		// Feature 0
		if (flag)
		{
			*pd1_2 = edge_fea_intensity - *pd1_2;				
			*pd1 += *pd1_2;
		}
		else
		{
			*pd1_2 = - *pd1_2;
			*pd1 += *pd1_2;
		}
		p_edge ++;
	}

	// 2nd, left end pair
	pd1 = *(pd2 ++) + 3;
	pd1_2 = *(pd2_2 ++) + 3;
	flag = true;

	sample1 = p_edge->p_node[1]->sampled_value;
	sample2 = p_edge->p_node[2]->sampled_value;

	// Feature 2
	if (sample1 == 3 && sample2 == 3)
	{
		flag = false;
		*pd1_2 = edge_fea_intensity - *pd1_2;				
		*(pd1--) += *(pd1_2 --);
	}
	else
	{
		*pd1_2 = - *pd1_2;
		*(pd1--) += *(pd1_2 --);
	}

	// Feature 1
	if (flag && (sample1 == 1 && sample2 == 3 || sample1 == 2 && sample2 == 3 ||
				sample1 == 3 && sample2 == 1 || sample1 == 3 && sample2 == 2))
	{
		flag = false;
		*pd1_2 = edge_fea_intensity - *pd1_2;
		*(pd1--) += *(pd1_2 --);
	}
	else
	{
		*pd1_2 = - *pd1_2;
		*(pd1--) += *(pd1_2 --);
	}

	// Feature 0
	if (flag)
	{
		*pd1_2 = edge_fea_intensity - *pd1_2;				
		*pd1 += *pd1_2;
	}
	else
	{
		*pd1_2 = - *pd1_2;
		*pd1 += *pd1_2;
	}
	p_edge ++;


	// 3rd: right end pair
	pd1 = *(pd2 ++) + 3;
	pd1_2 = *(pd2_2 ++) + 3;
	flag = true;

	sample1 = p_edge->p_node[1]->sampled_value;
	sample2 = p_edge->p_node[2]->sampled_value;

	// Feature 2
	if (sample1 == 2 && sample2 == 2)
	{
		flag = false;
		*pd1_2 = edge_fea_intensity - *pd1_2;				
		*(pd1--) += *(pd1_2 --);
	}
	else
	{
		*pd1_2 = - *pd1_2;
		*(pd1--) += *(pd1_2 --);
	}

	// Feature 1
	if (flag && (sample1 == 1 && sample2 == 2 || sample1 == 2 && sample2 == 1 ||
				sample1 == 2 && sample2 == 3 || sample1 == 3 && sample2 == 2))
	{
		flag = false;
		*pd1_2 = edge_fea_intensity - *pd1_2;				
		*(pd1--) += *(pd1_2 --);
	}
	else
	{
		*pd1_2 = - *pd1_2;
		*(pd1--) += *(pd1_2 --);
	}

	// Feature 0
	if (flag)
	{
		*pd1_2 = edge_fea_intensity - *pd1_2;				
		*pd1 += *pd1_2;
	}
	else
	{
		*pd1_2 = - *pd1_2;
		*pd1 += *pd1_2;
	}
}


/**
 *	Set the edge and node potentials of the graph.  
 *	Space allocation has been done by alloc_potentials()
 */
void CRFSensorCircle::set_potentials()
{
	int cur_item;
	double *cur_pot1, **nfp2, **efp2, dtemp;
	Node *p_node;
	Edge *p_edge;

	// Node features
	set_node_potentials();
	
	// Node Extra features
	p_node = node_array + 1;
	nfp2 = node_extra_fea_para + 1;

	for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
	{
		cur_pot1 = p_node->node_pot + 2;
		dtemp = *(*(nfp2 ++) + 1);
		*(cur_pot1 ++) += dtemp;
		*cur_pot1 += dtemp;

		p_node ++;
	}

	// Edge features
	p_edge = edge_array + 1;
	efp2 = edge_fea_para + 1;

	// 1st, upper and lower chain
	for (cur_item = 2; cur_item < global_num_edge; cur_item ++)
	{
		dtemp = edge_fea_intensity * (*efp2)[1];
		p_edge->edge_pot[1][1][1] += dtemp;
		p_edge->edge_pot[1][1][3] += dtemp;
		p_edge->edge_pot[1][2][1] += dtemp;
		p_edge->edge_pot[1][2][3] += dtemp;

		p_edge->edge_pot[2][1][1] += dtemp;
		p_edge->edge_pot[2][1][2] += dtemp;
		p_edge->edge_pot[2][3][1] += dtemp;
		p_edge->edge_pot[2][3][2] += dtemp;

		dtemp = edge_fea_intensity * (*efp2)[2];
		p_edge->edge_pot[1][1][2] += dtemp;
		p_edge->edge_pot[1][2][2] += dtemp;
		p_edge->edge_pot[1][3][1] += dtemp;
		p_edge->edge_pot[1][3][3] += dtemp;

		p_edge->edge_pot[2][1][3] += dtemp;
		p_edge->edge_pot[2][2][1] += dtemp;
		p_edge->edge_pot[2][2][2] += dtemp;
		p_edge->edge_pot[2][3][3] += dtemp;

		dtemp = edge_fea_intensity * (*efp2)[3];
		p_edge->edge_pot[1][3][2] += dtemp;
		p_edge->edge_pot[2][2][3] += dtemp;

		p_edge ++;	efp2 ++;			
	}

	// 2nd, left end pairs
	dtemp = edge_fea_intensity * (*efp2)[1];

	p_edge->edge_pot[1][1][1] += dtemp;
	p_edge->edge_pot[1][1][2] += dtemp;
	p_edge->edge_pot[1][2][1] += dtemp;
	p_edge->edge_pot[1][2][2] += dtemp;

	p_edge->edge_pot[2][1][1] += dtemp;
	p_edge->edge_pot[2][1][2] += dtemp;
	p_edge->edge_pot[2][2][1] += dtemp;
	p_edge->edge_pot[2][2][2] += dtemp;

	dtemp = edge_fea_intensity * (*efp2)[2];
	p_edge->edge_pot[1][1][3] += dtemp;
	p_edge->edge_pot[1][2][3] += dtemp;
	p_edge->edge_pot[1][3][1] += dtemp;
	p_edge->edge_pot[1][3][2] += dtemp;

	p_edge->edge_pot[2][1][3] += dtemp;
	p_edge->edge_pot[2][2][3] += dtemp;
	p_edge->edge_pot[2][3][1] += dtemp;
	p_edge->edge_pot[2][3][2] += dtemp;

	dtemp = edge_fea_intensity * (*efp2)[3];
	p_edge->edge_pot[1][3][3] += dtemp;
	p_edge->edge_pot[2][3][3] += dtemp;

	p_edge ++;	efp2 ++;	

	// 3rd, right end pairs
	dtemp = edge_fea_intensity * (*efp2)[1];

	p_edge->edge_pot[1][1][1] += dtemp;
	p_edge->edge_pot[1][1][3] += dtemp;
	p_edge->edge_pot[1][3][1] += dtemp;
	p_edge->edge_pot[1][3][3] += dtemp;

	p_edge->edge_pot[2][1][1] += dtemp;
	p_edge->edge_pot[2][1][3] += dtemp;
	p_edge->edge_pot[2][3][1] += dtemp;
	p_edge->edge_pot[2][3][3] += dtemp;

	dtemp = edge_fea_intensity * (*efp2)[2];
	p_edge->edge_pot[1][1][2] += dtemp;
	p_edge->edge_pot[1][2][1] += dtemp;
	p_edge->edge_pot[1][2][3] += dtemp;
	p_edge->edge_pot[1][3][2] += dtemp;

	p_edge->edge_pot[2][1][2] += dtemp;
	p_edge->edge_pot[2][2][1] += dtemp;
	p_edge->edge_pot[2][2][3] += dtemp;
	p_edge->edge_pot[2][3][2] += dtemp;

	dtemp = edge_fea_intensity * (*efp2)[3];
	p_edge->edge_pot[1][2][2] += dtemp;
	p_edge->edge_pot[2][2][2] += dtemp;
}


char * CRFSensorCircle::get_diagnosis(int argc, char ** argv)
{		
	return string_max_param;
}


}
