#include "TreeBasics.hh"


using namespace std;

namespace tree_mcmc {


/** 
 *	For CRF which is a chain, then work on this tree
 */

CRFToyTraffic::CRFToyTraffic(int *argc, char** argv) : CRFGrid(argc, argv)
{

}


/**
 *	After calculating the RB estimations on the graph,
 *	Calculate the feature expectations.
 */
void CRFToyTraffic::cal_fea_expectation()
{
	// Finally, calculate the node and edge feature expectations
	int cur_edge, cur_class;
	double *fept1, **fept2;
	Edge *p_edge;

	cal_node_fea_exp();

	// Second, calculate the edge feature expectation
	p_edge = edge_array + 1;
	fept2 = edge_fea_exp + 1;

	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{	
		fept1 = * (fept2 ++) + 1;
		*fept1 = 0.0;

/*		Most general and advanced case
		rb2 = p_edge->rb_est + 1;

		for (cur_class_1 = 1; cur_class_1 <= num_edge_fea; cur_class_1 ++)
		{
			rb1 = *(rb2++) + 1;
			for (cur_class_1 = 1; cur_class_1 <= num_edge_fea; cur_class_1 ++)
			{
//				edge_fea_ept[cur_edge][cur_fea]
//							+= edge_fea_fn_k(cur_class_1, cur_class_2) * p_edge->rb_est[cur_class_1][cur_class_2];

				*fept1 += edge_fea_fn(cur_fea, cur_class_1, cur_class_2) * (*rb1++);	
			}
		}
		Most general and advanced case
*/
		for (cur_class = 1; cur_class <= p_edge->p_node[1]->num_class; cur_class ++)
			 *fept1 += p_edge->rb_est[cur_class][cur_class];
		*fept1 *= edge_fea_intensity;

		p_edge ++;
	}
}


void CRFToyTraffic::fill_trace()
{
	int cur_item, cur_fea;
	double **pd2, *pd1, **pd2_2, *pd1_2;
	Edge *p_edge;

	fill_node_trace();

	pd2 = edge_fea_trace + 1;
	pd2_2 = edge_fea_exp + 1;
	p_edge = edge_array + 1;

	for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
	{
		pd1 = *(pd2 ++) + 1;
		pd1_2 = *(pd2_2 ++) + 1;

		if (p_edge->p_node[1]->sampled_value == p_edge->p_node[2]->sampled_value)
		{
			for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea++)
			{
				*pd1_2 = edge_fea_intensity - *pd1_2;				
				*(pd1 ++) += *(pd1_2 ++);
			}
		}
		else
		{	
			for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea++)
			{
				*pd1_2 = - *pd1_2;
				*(pd1++) += *(pd1_2 ++);				
			}
		}
		p_edge ++;
	}
}


/**
 *	Set the edge and node potentials of the graph.  
 *	Space allocation has been done by alloc_potentials()
 */
void CRFToyTraffic::set_potentials()
{
	int cur_item, cur_class;
	Edge *p_edge;
	double **efp2;

	set_node_potentials();

	//////////////////////////////////////////////////////////////////////////
	//  Edge potentials

	p_edge = edge_array + 1;
	efp2 = edge_fea_para + 1;

	for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)		// loop over edges
	{
		for (cur_class = 1; cur_class <= num_class; cur_class ++)
		{
			p_edge->edge_pot[1][cur_class][cur_class] = edge_fea_intensity * (*efp2)[1];
			p_edge->edge_pot[2][cur_class][cur_class] = p_edge->edge_pot[1][cur_class][cur_class];
		}

		p_edge ++;	efp2 ++;
	}
}


char * CRFToyTraffic::get_diagnosis(int argc, char ** argv)
{		
// 	if (rand() < RAND_MAX / 1000.0)
//  	{
//  		dump_graph("spec.dat", "my_result.dat");
//  		cin.get();
//  	}
	return string_max_param;

}

}
