#ifndef Utility_hh
#define Utility_hh

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "inttypes.h"

	
//#define i386
typedef union
{
    double d;
    int64_t l;
    struct
    {
#if defined(i386) || defined(__alpha)
/* #ifdef __LITTLE_ENDIAN */
        int32_t j, i;
#else
        int32_t i, j;
#endif
    } n;
    struct
    {
#if defined(i386) || defined(__alpha)
/* #ifdef __LITTLE_ENDIAN */
        uint32_t j, i;
#else
        uint32_t i, j;
#endif
    } u;
} _eco;



class Utility {

public:

/**
 *	Extract an integer from arguments
 *	mode = 0 allows para not to exist in argv, and then return -1
 *	mode = 1 forces para to exist in argv, otherwise terminate
 *	mode = 2 when not exist, then add to argc, argv, and then return 0
 *	Return 1 if found the parameter, 0 if not found
 */
static int get_arg(int *result, int *argc, char **argv, char *para, int mode, int def_val);


/**
 *	Extract an unsigned integer from arguments
 *	mode = 0 allows para not to exist in argv, and then return -1
 *	mode = 1 forces para to exist in argv, otherwise terminate
 *	mode = 2 when not exist, then add to argc, argv, and then return 0
 *	Return 1 if found the parameter, 0 if not found
 */
static int get_arg(unsigned int *result, int *argc, char **argv, char *para, int mode, unsigned int def_val);


/**
 *	Extract an boolean from arguments
 *	mode = 0 allows para not to exist in argv, and then return -1
 *	mode = 1 forces para to exist in argv, otherwise terminate
 *	mode = 2 when not exist, then add to argc, argv, and then return 0
 *	Return 1 if found the parameter, 0 if not found
 */
static int get_arg(bool *result, int *argc, char ** argv, char *para, int mode, bool def_val);


/**
 *	Extract a long integer from arguments
 *	mode = 0 allows para not to exist in argv, and then return -1
 *	mode = 1 forces para to exist in argv, otherwise terminate
 *	mode = 2 when not exist, then add to argc, argv, and then return 0
 *	Return 1 if found the parameter, 0 if not found
 */
static int get_arg(long *result, int *argc, char **argv, char *para, int mode, long def_val);


/**
 *	Extract a double precision floating point number from arguments
 *	mode = 0 allows para not to exist in argv, and then return -1
 *	mode = 1 forces para to exist in argv, otherwise terminate
 *	mode = 2 when not exist, then add to argc, argv, and then return 0
 *	Return 1 if found the parameter, 0 if not found
 */
static int get_arg(double *result, int *argc, char **argv, char *para, int mode, double def_val);


/**
 *	Extract a string from arguments
 *	mode = 0 allows para not to exist in argv, and then return -1
 *	mode = 1 forces para to exist in argv, otherwise terminate
 *	mode = 2 when not exist, then add to argc, argv, and then return 0
 *	Return 1 if found the parameter, 0 if not found
 */
static int get_arg(char **result, int *argc, char **argv, char *para, int mode, char *def_val);


/*
 *	Copy _argc and _argv to argc and argv
 */
static void copy_arg(int * argc, char ** argv, int _argc, char ** _argv);


/*
 *	Overridden functions that add a (flag, argument) pair to argc and argv
 */
static void add_arg(int * argc, char ** argv, int value, char *flag);
static void add_arg(int * argc, char ** argv, unsigned int value, char *flag);
static void add_arg(int * argc, char ** argv, bool value, char *flag);
static void add_arg(int * argc, char ** argv, long value, char *flag);
static void add_arg(int * argc, char ** argv, char * buffer, char *flag);
static void add_arg(int * argc, char ** argv, double value, char *flag);

static void destroy_all_args(int argc, char **argv);


static double Utility::fexp2(register double y);
static double Utility::fexp2l(register double y);
static double Utility::flog2(register double y);
static double Utility::flog2l(register double y);
static double Utility::fexp2r(register double y);
static double Utility::fexp2r_check(register double y);
static double Utility::fexp2rl(register double y);
static double Utility::flog2r(register double y);
static double Utility::flog2rl(register double y);

};


inline double Utility::fexp2(register double y)
{
    register _eco tmp;

    tmp.d = y;
    tmp.n.i += (20 << 20);
    tmp.n.i = (int32_t) tmp.d;
    tmp.n.i += (1023 << 20);
    tmp.n.j = 0;
    return tmp.d;
}

inline double Utility::fexp2l(register double y)
{
    register _eco tmp;

    tmp.d = y;
    tmp.n.i += (52 << 20);
    tmp.l = (int64_t) tmp.d;
    tmp.n.i += (1023 << 20);
    return tmp.d;
}

inline double Utility::flog2(register double y)
{
    register _eco tmp;

    tmp.d = y;
    tmp.n.i -= (1023 << 20);
    if (tmp.n.i)
    {
        tmp.d = (double) tmp.n.i;
        tmp.n.i -= (20 << 20);
    }
    return tmp.d;
}

inline double Utility::flog2l(register double y)
{
    register _eco tmp;

    tmp.d = y;
    tmp.n.i -= (1023 << 20);
    if (tmp.l)
    {
        tmp.d = (double) tmp.l;
        tmp.n.i -= (52 << 20);
    }
    return tmp.d;
}

/* EXP2(y) = EXP(y/2)/EXP(-y/2) */ 

inline double Utility::fexp2r(register double y)
{
    register _eco tmp, tmp2;

    tmp.d = y;
    tmp.n.i += (19 << 20);
    tmp2.n.i = (1023 << 20) - (int32_t) tmp.d;
    tmp.n.i = (1023 << 21) - tmp2.n.i;
    tmp.n.j = tmp2.n.j = 0;
    return tmp.d/tmp2.d;
}


inline double Utility::fexp2r_check(register double y)
{
    register _eco tmp, tmp2;

	if (y > 1010)		// largest acceptable exponent (base 2)
	{
		printf("Too large exponent: %f\n", y);
		exit(1);
	}
	else if (y < -1010)
		return 0.0;

    tmp.d = y;
    tmp.n.i += (19 << 20);
    tmp2.n.i = (1023 << 20) - (int32_t) tmp.d;
    tmp.n.i = (1023 << 21) - tmp2.n.i;
    tmp.n.j = tmp2.n.j = 0;
    return tmp.d/tmp2.d;
}


inline double Utility::fexp2rl(register double y)
{
    register _eco tmp, tmp2;

    tmp.d = y;
    tmp.n.i += (51 << 20);
    tmp.l = (int64_t) tmp.d;
    tmp2.l = -tmp.l;
    tmp.n.i += (1023 << 20);
    tmp2.n.i += (1023 << 20);
    return tmp.d/tmp2.d;
}

/* LOG2(y) = (LOG(y) - LOG(1/y))/2 */

inline double Utility::flog2r(register double y)
{
    register _eco tmp, tmp2;

    tmp.d = y;
    tmp2.d = 1.0/y;
    tmp.n.i -= tmp2.n.i;
    if (tmp.n.i)
    {
        tmp.d = (double) tmp.n.i;
        tmp.n.i -= (21 << 20);
    }
    return tmp.d;
}

inline double Utility::flog2rl(register double y)
{
    register _eco tmp, tmp2;

    tmp.d = y;
    tmp2.d = 1.0/y;
    tmp.l -= tmp2.l;
    if (tmp.l)
    {
        tmp.d = (double) tmp.l;
        tmp.n.i -= (21 << 20);
    }
    return tmp.d;
}


#endif
